@@ -19,10 +19,14 @@ use itertools::Itertools;
 use tracing::instrument;
 
 use crate::cli_util::{
-    get_new_config_file_path, run_ui_editor, serialize_config_value, user_error,
-    write_config_value_to_file, CommandError, CommandHelper,
+    get_new_config_file_path, run_ui_editor, serialize_config_value, write_config_value_to_file,
+    CommandHelper,
 };
+use crate::command_error::{config_error, user_error, CommandError};
 use crate::config::{AnnotatedValue, ConfigSource};
+use crate::generic_templater::GenericTemplateLanguage;
+use crate::template_builder::TemplateLanguage as _;
+use crate::templater::TemplatePropertyExt as _;
 use crate::ui::Ui;
 
 #[derive(clap::Args, Clone, Debug)]
@@ -91,6 +95,17 @@ pub(crate) struct ConfigListArgs {
     #[arg(long)]
     repo: bool,
     // TODO(#1047): Support --show-origin using LayeredConfigs.
+    /// Render each variable using the given template
+    ///
+    /// The following keywords are defined:
+    ///
+    /// * `name: String`: Config name.
+    /// * `value: String`: Serialized value in TOML syntax.
+    /// * `overridden: Boolean`: True if the value is shadowed by other.
+    ///
+    /// For the syntax, see https://github.com/martinvonz/jj/blob/main/docs/templates.md
+    #[arg(long, short = 'T', verbatim_doc_comment)]
+    template: Option<String>,
 }
 
 impl ConfigListArgs {
@@ -169,56 +184,82 @@ pub(crate) fn cmd_config(
     }
 }
 
+// AnnotatedValue will be cloned internally in the templater. If the cloning
+// cost matters, wrap it with Rc.
+fn config_template_language() -> GenericTemplateLanguage<'static, AnnotatedValue> {
+    type L = GenericTemplateLanguage<'static, AnnotatedValue>;
+    let mut language = L::new();
+    // "name" instead of "path" to avoid confusion with the source file path
+    language.add_keyword("name", |self_property| {
+        let out_property = self_property.map(|annotated| annotated.path.join("."));
+        Ok(L::wrap_string(out_property))
+    });
+    language.add_keyword("value", |self_property| {
+        // TODO: would be nice if we can provide raw dynamically-typed value
+        let out_property = self_property.map(|annotated| serialize_config_value(&annotated.value));
+        Ok(L::wrap_string(out_property))
+    });
+    language.add_keyword("overridden", |self_property| {
+        let out_property = self_property.map(|annotated| annotated.is_overridden);
+        Ok(L::wrap_boolean(out_property))
+    });
+    language
+}
+
 #[instrument(skip_all)]
 pub(crate) fn cmd_config_list(
     ui: &mut Ui,
     command: &CommandHelper,
     args: &ConfigListArgs,
 ) -> Result<(), CommandError> {
+    let template = {
+        let language = config_template_language();
+        let text = match &args.template {
+            Some(value) => value.to_owned(),
+            None => command
+                .settings()
+                .config()
+                .get_string("templates.config_list")?,
+        };
+        command.parse_template(ui, &language, &text, GenericTemplateLanguage::wrap_self)?
+    };
+
     ui.request_pager();
+    let mut formatter = ui.stdout_formatter();
+    formatter.push_label("config_list")?;
     let name_path = args
         .name
         .as_ref()
         .map_or(vec![], |name| name.split('.').collect_vec());
-    let values = command.resolved_config_values(&name_path)?;
     let mut wrote_values = false;
-    for AnnotatedValue {
-        path,
-        value,
-        source,
-        is_overridden,
-    } in &values
-    {
+    for annotated in command.resolved_config_values(&name_path)? {
         // Remove overridden values.
-        if *is_overridden && !args.include_overridden {
+        if annotated.is_overridden && !args.include_overridden {
             continue;
         }
 
         if let Some(target_source) = args.get_source_kind() {
-            if target_source != *source {
+            if target_source != annotated.source {
                 continue;
             }
         }
 
         // Skip built-ins if not included.
-        if !args.include_defaults && *source == ConfigSource::Default {
+        if !args.include_defaults && annotated.source == ConfigSource::Default {
             continue;
         }
-        writeln!(
-            ui.stdout(),
-            "{}{}={}",
-            if *is_overridden { "# " } else { "" },
-            path.join("."),
-            serialize_config_value(value)
-        )?;
+
+        template.format(&annotated, formatter.as_mut())?;
         wrote_values = true;
     }
+    formatter.pop_label()?;
+    drop(formatter);
     if !wrote_values {
         // Note to stderr explaining why output is empty.
         if let Some(name) = &args.name {
-            writeln!(ui.warning(), "No matching config key for {name}")?;
+            writeln!(ui.warning_default(), "No matching config key for {name}")?;
         } else {
-            writeln!(ui.warning(), "No config to list")?;
+            writeln!(ui.warning_default(), "No config to list")?;
         }
     }
     Ok(())
@@ -254,7 +295,7 @@ pub(crate) fn cmd_config_get(
                 if let Some(origin) = origin {
                     write!(buf, " in {origin}").unwrap();
                 }
-                CommandError::ConfigError(buf.to_string())
+                config_error(buf)
             }
             err => err.into(),
         })?;
