@@ -19,8 +19,10 @@ use jj_lib::rewrite::rebase_to_dest_parent;
 use tracing::instrument;
 
 use crate::cli_util::{
-    CommandError, CommandHelper, LogContentFormat, RevisionArg, WorkspaceCommandHelper,
+    format_template, CommandHelper, LogContentFormat, RevisionArg, WorkspaceCommandHelper,
 };
+use crate::command_error::CommandError;
+use crate::commit_templater::CommitTemplateLanguage;
 use crate::diff_util::{self, DiffFormat, DiffFormatArgs};
 use crate::formatter::Formatter;
 use crate::graphlog::{get_graphlog, Edge};
@@ -63,19 +65,32 @@ pub(crate) fn cmd_obslog(
 ) -> Result<(), CommandError> {
     let workspace_command = command.workspace_helper(ui)?;
 
-    let start_commit = workspace_command.resolve_single_rev(&args.revision, ui)?;
-    let wc_commit_id = workspace_command.get_wc_commit_id();
+    let start_commit = workspace_command.resolve_single_rev(&args.revision)?;
 
     let diff_formats =
         diff_util::diff_formats_for_log(command.settings(), &args.diff_format, args.patch)?;
-
-    let template_string = match &args.template {
-        Some(value) => value.to_string(),
-        None => command.settings().config().get_string("templates.log")?,
-    };
-    let template = workspace_command.parse_commit_template(&template_string)?;
     let with_content_format = LogContentFormat::new(ui, command.settings())?;
 
+    let template;
+    let node_template;
+    {
+        let language = workspace_command.commit_template_language()?;
+        let template_string = match &args.template {
+            Some(value) => value.to_string(),
+            None => command.settings().config().get_string("templates.log")?,
+        };
+        template = workspace_command.parse_template(
+            &language,
+            &template_string,
+            CommitTemplateLanguage::wrap_commit,
+        )?;
+        node_template = workspace_command.parse_template(
+            &language,
+            &command.settings().commit_node_template(),
+            CommitTemplateLanguage::wrap_commit_opt,
+        )?;
+    }
+
     ui.request_pager();
     let mut formatter = ui.stdout_formatter();
     let formatter = formatter.as_mut();
@@ -91,11 +106,10 @@ pub(crate) fn cmd_obslog(
     }
     if !args.no_graph {
         let mut graph = get_graphlog(command.settings(), formatter.raw());
-        let default_node_symbol = graph.default_node_symbol().to_owned();
         for commit in commits {
             let mut edges = vec![];
             for predecessor in &commit.predecessors() {
-                edges.push(Edge::direct(predecessor.id().clone()));
+                edges.push(Edge::Direct(predecessor.id().clone()));
             }
             let mut buffer = vec![];
             with_content_format.write_graph_text(
@@ -116,15 +130,11 @@ pub(crate) fn cmd_obslog(
                     &diff_formats,
                 )?;
             }
-            let node_symbol = if Some(commit.id()) == wc_commit_id {
-                "@"
-            } else {
-                &default_node_symbol
-            };
+            let node_symbol = format_template(ui, &Some(commit.clone()), &node_template);
             graph.add_node(
                 commit.id(),
                 &edges,
-                node_symbol,
+                &node_symbol,
                 &String::from_utf8_lossy(&buffer),
             )?;
         }
