@@ -16,11 +16,12 @@ use itertools::Itertools;
 use jj_lib::repo::Repo;
 use jj_lib::revset::{RevsetExpression, RevsetIteratorExt};
 
-use crate::cli_util::{short_commit_hash, user_error, CommandError, CommandHelper};
+use crate::cli_util::{short_commit_hash, CommandHelper};
+use crate::command_error::{user_error, CommandError};
 use crate::commands::next::choose_commit;
 use crate::ui::Ui;
 
-/// Move the working copy commit to the parent of the current revision.
+/// Move the working-copy commit to the parent revision
 ///
 ///
 /// The command moves you to the parent in a linear fashion.
@@ -29,7 +30,7 @@ use crate::ui::Ui;
 /// D @  D
 /// |/   |
 /// A => A @
-/// |    | /
+/// |    |/
 /// B    B
 /// ```
 ///
@@ -44,6 +45,9 @@ use crate::ui::Ui;
 /// B    B
 /// |    |
 /// A    A
+///
+/// If your working-copy commit already has visible children, then `--edit` is
+/// implied.
 /// ```
 // TODO(#2126): Handle multiple parents, e.g merges.
 #[derive(clap::Args, Clone, Debug)]
@@ -63,13 +67,17 @@ pub(crate) fn cmd_prev(
     args: &PrevArgs,
 ) -> Result<(), CommandError> {
     let mut workspace_command = command.workspace_helper(ui)?;
-    let edit = args.edit;
     let amount = args.amount;
     let current_wc_id = workspace_command
         .get_wc_commit_id()
         .ok_or_else(|| user_error("This command requires a working copy"))?;
+    let edit = args.edit
+        || !workspace_command
+            .repo()
+            .view()
+            .heads()
+            .contains(current_wc_id);
     let current_wc = workspace_command.repo().store().get_commit(current_wc_id)?;
-    let current_short = short_commit_hash(current_wc.id());
     let start_id = if edit {
         current_wc_id
     } else {
@@ -93,7 +101,6 @@ pub(crate) fn cmd_prev(
         .evaluate_programmatic(workspace_command.repo().as_ref())?
         .iter()
         .commits(workspace_command.repo().store())
-        .take(2)
         .try_collect()?;
     let target = match targets.as_slice() {
         [target] => target,
@@ -106,6 +113,7 @@ pub(crate) fn cmd_prev(
         commits => choose_commit(ui, &workspace_command, "prev", commits)?,
     };
     // Generate a short commit hash, to make it readable in the op log.
+    let current_short = short_commit_hash(current_wc.id());
     let target_short = short_commit_hash(target.id());
     // If we're editing, just move to the revision directly.
     if edit {
