@@ -12,12 +12,13 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use jj_lib::backend::ObjectId;
 use jj_lib::matchers::EverythingMatcher;
+use jj_lib::object_id::ObjectId;
 use jj_lib::rewrite::merge_commit_trees;
 use tracing::instrument;
 
-use crate::cli_util::{user_error, CommandError, CommandHelper, RevisionArg};
+use crate::cli_util::{CommandHelper, RevisionArg};
+use crate::command_error::{user_error, CommandError};
 use crate::description_util::combine_messages;
 use crate::ui::Ui;
 
@@ -31,8 +32,10 @@ use crate::ui::Ui;
 /// If the source became empty and both the source and destination had a
 /// non-empty description, you will be asked for the combined description. If
 /// either was empty, then the other one will be used.
+///
+/// If a working-copy commit gets abandoned, it will be given a new, empty
+/// commit. This is true in general; it is not specific to this command.
 #[derive(clap::Args, Clone, Debug)]
-#[command(visible_alias = "unamend")]
 pub(crate) struct UnsquashArgs {
     #[arg(long, short, default_value = "@")]
     revision: RevisionArg,
@@ -41,6 +44,9 @@ pub(crate) struct UnsquashArgs {
     // the default.
     #[arg(long, short)]
     interactive: bool,
+    /// Specify diff editor to be used (implies --interactive)
+    #[arg(long, value_name = "NAME")]
+    tool: Option<String>,
 }
 
 #[instrument(skip_all)]
@@ -50,7 +56,7 @@ pub(crate) fn cmd_unsquash(
     args: &UnsquashArgs,
 ) -> Result<(), CommandError> {
     let mut workspace_command = command.workspace_helper(ui)?;
-    let commit = workspace_command.resolve_single_rev(&args.revision, ui)?;
+    let commit = workspace_command.resolve_single_rev(&args.revision)?;
     workspace_command.check_rewritable([&commit])?;
     let parents = commit.parents();
     if parents.len() != 1 {
@@ -58,10 +64,15 @@ pub(crate) fn cmd_unsquash(
     }
     let parent = &parents[0];
     workspace_command.check_rewritable(&parents[..1])?;
+    let interactive_editor = if args.tool.is_some() || args.interactive {
+        Some(workspace_command.diff_editor(ui, args.tool.as_deref())?)
+    } else {
+        None
+    };
     let mut tx = workspace_command.start_transaction();
     let parent_base_tree = merge_commit_trees(tx.repo(), &parent.parents())?;
     let new_parent_tree_id;
-    if args.interactive {
+    if let Some(diff_editor) = &interactive_editor {
         let instructions = format!(
             "\
 You are moving changes from: {}
@@ -78,12 +89,11 @@ aborted.
             tx.format_commit_summary(&commit)
         );
         let parent_tree = parent.tree()?;
-        new_parent_tree_id = tx.edit_diff(
-            ui,
+        new_parent_tree_id = diff_editor.edit(
             &parent_base_tree,
             &parent_tree,
             &EverythingMatcher,
-            &instructions,
+            Some(&instructions),
         )?;
         if new_parent_tree_id == parent_base_tree.id() {
             return Err(user_error("No changes selected"));
@@ -95,8 +105,7 @@ aborted.
     // case).
     if new_parent_tree_id == parent_base_tree.id() {
         tx.mut_repo().record_abandoned_commit(parent.id().clone());
-        let description =
-            combine_messages(tx.base_repo(), parent, &commit, command.settings(), true)?;
+        let description = combine_messages(tx.base_repo(), &[parent], &commit, command.settings())?;
         // Commit the new child on top of the parent's parents.
         tx.mut_repo()
             .rewrite_commit(command.settings(), &commit)
