@@ -15,21 +15,24 @@
 use std::fmt::Debug;
 use std::fs;
 use std::io::Write;
+use std::sync::Arc;
 
 use clap::Subcommand;
 use itertools::Itertools;
-use jj_lib::backend::ObjectId;
 use jj_lib::file_util;
-use jj_lib::op_store::WorkspaceId;
-use jj_lib::repo::Repo;
+use jj_lib::object_id::ObjectId;
+use jj_lib::op_store::{OpStoreError, WorkspaceId};
+use jj_lib::operation::Operation;
+use jj_lib::repo::{ReadonlyRepo, Repo};
 use jj_lib::rewrite::merge_commit_trees;
-use jj_lib::workspace::{default_working_copy_initializer, Workspace};
+use jj_lib::workspace::Workspace;
 use tracing::instrument;
 
 use crate::cli_util::{
-    self, check_stale_working_copy, print_checkout_stats, user_error, CommandError, CommandHelper,
-    RevisionArg, WorkspaceCommandHelper,
+    check_stale_working_copy, print_checkout_stats, short_commit_hash, CommandHelper, RevisionArg,
+    WorkingCopyFreshness, WorkspaceCommandHelper,
 };
+use crate::command_error::{internal_error_with_message, user_error, CommandError};
 use crate::ui::Ui;
 
 /// Commands for working with workspaces
@@ -41,6 +44,8 @@ use crate::ui::Ui;
 /// Each workspace has its own working-copy commit. When you have more than one
 /// workspace attached to a repo, they are indicated by `@<workspace name>` in
 /// `jj log`.
+///
+/// Each workspace also has own sparse patterns.
 #[derive(Subcommand, Clone, Debug)]
 pub(crate) enum WorkspaceCommand {
     Add(WorkspaceAddArgs),
@@ -51,6 +56,8 @@ pub(crate) enum WorkspaceCommand {
 }
 
 /// Add a workspace
+///
+/// Sparse patterns will be copied over from the current workspace.
 #[derive(clap::Args, Clone, Debug)]
 pub(crate) struct WorkspaceAddArgs {
     /// Where to create the new workspace
@@ -148,22 +155,36 @@ fn cmd_workspace_add(
             "Workspace named '{name}' already exists"
         )));
     }
-    // TODO: How do we create a workspace with a non-default working copy?
+
+    let working_copy_factory = command.get_working_copy_factory()?;
     let (new_workspace, repo) = Workspace::init_workspace_with_existing_repo(
         command.settings(),
         &destination_path,
         repo,
-        default_working_copy_initializer(),
+        working_copy_factory,
         workspace_id,
     )?;
     writeln!(
-        ui.stderr(),
+        ui.status(),
         "Created workspace in \"{}\"",
         file_util::relative_path(old_workspace_command.workspace_root(), &destination_path)
             .display()
     )?;
 
+    // Copy sparse patterns from workspace where the command was run
     let mut new_workspace_command = WorkspaceCommandHelper::new(ui, command, new_workspace, repo)?;
+    let (mut locked_ws, _wc_commit) = new_workspace_command.start_working_copy_mutation()?;
+    let sparse_patterns = old_workspace_command
+        .working_copy()
+        .sparse_patterns()?
+        .to_vec();
+    locked_ws
+        .locked_wc()
+        .set_sparse_patterns(sparse_patterns)
+        .map_err(|err| internal_error_with_message("Failed to set sparse patterns", err))?;
+    let operation_id = locked_ws.locked_wc().old_operation_id().clone();
+    locked_ws.finish(operation_id)?;
+
     let mut tx = new_workspace_command.start_transaction();
 
     // If no parent revisions are specified, create a working-copy commit based
@@ -181,7 +202,8 @@ fn cmd_workspace_add(
             vec![tx.repo().store().root_commit()]
         }
     } else {
-        cli_util::resolve_all_revs(&old_workspace_command, ui, &args.revision)?
+        old_workspace_command
+            .resolve_some_revsets_default_single(&args.revision)?
             .into_iter()
             .collect_vec()
     };
@@ -254,11 +276,13 @@ fn cmd_workspace_list(
 ) -> Result<(), CommandError> {
     let workspace_command = command.workspace_helper(ui)?;
     let repo = workspace_command.repo();
+    let mut formatter = ui.stdout_formatter();
+    let template = workspace_command.commit_summary_template();
     for (workspace_id, wc_commit_id) in repo.view().wc_commit_ids().iter().sorted() {
-        write!(ui.stdout(), "{}: ", workspace_id.as_str())?;
+        write!(formatter, "{}: ", workspace_id.as_str())?;
         let commit = repo.store().get_commit(wc_commit_id)?;
-        workspace_command.write_commit_summary(ui.stdout_formatter().as_mut(), &commit)?;
-        writeln!(ui.stdout())?;
+        template.format(&commit, formatter.as_mut())?;
+        writeln!(formatter)?;
     }
     Ok(())
 }
@@ -269,8 +293,8 @@ fn cmd_workspace_root(
     command: &CommandHelper,
     _args: &WorkspaceRootArgs,
 ) -> Result<(), CommandError> {
-    let workspace_command = command.workspace_helper(ui)?;
-    let root = workspace_command
+    let root = command
+        .workspace_loader()?
         .workspace_root()
         .to_str()
         .ok_or_else(|| user_error("The workspace root is not valid UTF-8"))?;
@@ -278,6 +302,73 @@ fn cmd_workspace_root(
     Ok(())
 }
 
+fn create_and_check_out_recovery_commit(
+    ui: &mut Ui,
+    command: &CommandHelper,
+) -> Result<Arc<ReadonlyRepo>, CommandError> {
+    let mut workspace_command = command.workspace_helper_no_snapshot(ui)?;
+    let workspace_id = workspace_command.workspace_id().clone();
+    let mut tx = workspace_command.start_transaction().into_inner();
+
+    let (mut locked_workspace, commit) =
+        workspace_command.unchecked_start_working_copy_mutation()?;
+    let commit_id = commit.id();
+
+    let mut_repo = tx.mut_repo();
+    let new_commit = mut_repo
+        .new_commit(
+            command.settings(),
+            vec![commit_id.clone()],
+            commit.tree_id().clone(),
+        )
+        .write()?;
+    mut_repo.set_wc_commit(workspace_id, new_commit.id().clone())?;
+    let repo = tx.commit("recovery commit");
+
+    locked_workspace.locked_wc().recover(&new_commit)?;
+    locked_workspace.finish(repo.op_id().clone())?;
+
+    writeln!(
+        ui.status(),
+        "Created and checked out recovery commit {}",
+        short_commit_hash(new_commit.id())
+    )?;
+
+    Ok(repo)
+}
+
+/// Loads workspace that will diverge from the last working-copy operation.
+fn for_stale_working_copy(
+    ui: &mut Ui,
+    command: &CommandHelper,
+) -> Result<(WorkspaceCommandHelper, bool), CommandError> {
+    let workspace = command.load_workspace()?;
+    let op_store = workspace.repo_loader().op_store();
+    let (repo, recovered) = {
+        let op_id = workspace.working_copy().operation_id();
+        match op_store.read_operation(op_id) {
+            Ok(op_data) => (
+                workspace.repo_loader().load_at(&Operation::new(
+                    op_store.clone(),
+                    op_id.clone(),
+                    op_data,
+                ))?,
+                false,
+            ),
+            Err(e @ OpStoreError::ObjectNotFound { .. }) => {
+                writeln!(
+                    ui.status(),
+                    "Failed to read working copy's current operation; attempting recovery. Error \
+                     message from read attempt: {e}"
+                )?;
+                (create_and_check_out_recovery_commit(ui, command)?, true)
+            }
+            Err(e) => return Err(e.into()),
+        }
+    };
+    Ok((command.for_loaded_repo(ui, workspace, repo)?, recovered))
+}
+
 #[instrument(skip_all)]
 fn cmd_workspace_update_stale(
     ui: &mut Ui,
@@ -289,8 +380,16 @@ fn cmd_workspace_update_stale(
     // merged repo wouldn't change because the old one wins, but it's probably
     // fine if we picked the new wc_commit_id.
     let known_wc_commit = {
-        let mut workspace_command = command.for_stale_working_copy(ui)?;
-        workspace_command.snapshot(ui)?;
+        let (mut workspace_command, recovered) = for_stale_working_copy(ui, command)?;
+        workspace_command.maybe_snapshot(ui)?;
+
+        if recovered {
+            // We have already recovered from the situation that prompted the user to run
+            // this command, and it is known that the workspace is not stale
+            // (since we just updated it), so we can return early.
+            return Ok(());
+        }
+
         let wc_commit_id = workspace_command.get_wc_commit_id().unwrap();
         workspace_command.repo().store().get_commit(wc_commit_id)?
     };
@@ -299,34 +398,41 @@ fn cmd_workspace_update_stale(
     let repo = workspace_command.repo().clone();
     let (mut locked_ws, desired_wc_commit) =
         workspace_command.unchecked_start_working_copy_mutation()?;
-    if !check_stale_working_copy(locked_ws.locked_wc(), &desired_wc_commit, &repo)?.is_stale() {
-        writeln!(
-            ui.stderr(),
-            "Nothing to do (the working copy is not stale)."
-        )?;
-    } else {
-        // The same check as start_working_copy_mutation(), but with the stale
-        // working-copy commit.
-        if known_wc_commit.tree_id() != locked_ws.locked_wc().old_tree_id() {
-            return Err(user_error("Concurrent working copy operation. Try again."));
+    match check_stale_working_copy(locked_ws.locked_wc(), &desired_wc_commit, &repo)? {
+        WorkingCopyFreshness::Fresh | WorkingCopyFreshness::Updated(_) => {
+            writeln!(
+                ui.status(),
+                "Nothing to do (the working copy is not stale)."
+            )?;
+        }
+        WorkingCopyFreshness::WorkingCopyStale | WorkingCopyFreshness::SiblingOperation => {
+            // The same check as start_working_copy_mutation(), but with the stale
+            // working-copy commit.
+            if known_wc_commit.tree_id() != locked_ws.locked_wc().old_tree_id() {
+                return Err(user_error("Concurrent working copy operation. Try again."));
+            }
+            let stats = locked_ws
+                .locked_wc()
+                .check_out(&desired_wc_commit)
+                .map_err(|err| {
+                    internal_error_with_message(
+                        format!(
+                            "Failed to check out commit {}",
+                            desired_wc_commit.id().hex()
+                        ),
+                        err,
+                    )
+                })?;
+            locked_ws.finish(repo.op_id().clone())?;
+            if let Some(mut formatter) = ui.status_formatter() {
+                write!(formatter, "Working copy now at: ")?;
+                formatter.with_label("working_copy", |fmt| {
+                    workspace_command.write_commit_summary(fmt, &desired_wc_commit)
+                })?;
+                writeln!(formatter)?;
+            }
+            print_checkout_stats(ui, stats, &desired_wc_commit)?;
         }
-        let stats = locked_ws
-            .locked_wc()
-            .check_out(&desired_wc_commit)
-            .map_err(|err| {
-                CommandError::InternalError(format!(
-                    "Failed to check out commit {}: {}",
-                    desired_wc_commit.id().hex(),
-                    err
-                ))
-            })?;
-        locked_ws.finish(repo.op_id().clone())?;
-        write!(ui.stderr(), "Working copy now at: ")?;
-        ui.stderr_formatter().with_label("working_copy", |fmt| {
-            workspace_command.write_commit_summary(fmt, &desired_wc_commit)
-        })?;
-        writeln!(ui.stderr())?;
-        print_checkout_stats(ui, stats, &desired_wc_commit)?;
     }
     Ok(())
 }
