@@ -97,12 +97,13 @@ impl TestEnvironment {
         cmd.env("JJ_EMAIL", "test.user@example.com");
         cmd.env("JJ_OP_HOSTNAME", "host.example.com");
         cmd.env("JJ_OP_USERNAME", "test-username");
+        cmd.env("JJ_TZ_OFFSET_MINS", "660");
 
         let mut command_number = self.command_number.borrow_mut();
         *command_number += 1;
         cmd.env("JJ_RANDOMNESS_SEED", command_number.to_string());
         let timestamp = chrono::DateTime::parse_from_rfc3339("2001-02-03T04:05:06+07:00").unwrap();
-        let timestamp = timestamp + chrono::Duration::seconds(*command_number);
+        let timestamp = timestamp + chrono::Duration::try_seconds(*command_number).unwrap();
         cmd.env("JJ_TIMESTAMP", timestamp.to_rfc3339());
         cmd.env("JJ_OP_TIMESTAMP", timestamp.to_rfc3339());
 
@@ -114,27 +115,29 @@ impl TestEnvironment {
             cmd.env("SSL_CERT_FILE", "/dev/null");
         }
 
-        if cfg!(all(windows, target_env = "gnu")) {
-            // MinGW executables cannot run without `mingw\bin` in the PATH (which we're
-            // clearing above), so we add it again here.
-            if let Ok(path_var) = std::env::var("PATH").or_else(|_| std::env::var("Path")) {
-                // There can be slight variations of this path (e.g. `mingw64\bin`) so we're
-                // intentionally being lenient here.
-                let mingw_directories = path_var
-                    .split(';')
-                    .filter(|dir| dir.contains("mingw"))
-                    .join(";");
-
-                if !mingw_directories.is_empty() {
-                    cmd.env("PATH", mingw_directories);
-                }
-            }
-
-            // MinGW uses `TEMP` to create temporary directories, which we need for some
+        if cfg!(windows) {
+            // Windows uses `TEMP` to create temporary directories, which we need for some
             // tests.
             if let Ok(tmp_var) = std::env::var("TEMP") {
                 cmd.env("TEMP", tmp_var);
             }
+
+            if cfg!(target_env = "gnu") {
+                // MinGW executables cannot run without `mingw\bin` in the PATH (which we're
+                // clearing above), so we add it again here.
+                if let Ok(path_var) = std::env::var("PATH").or_else(|_| std::env::var("Path")) {
+                    // There can be slight variations of this path (e.g. `mingw64\bin`) so we're
+                    // intentionally being lenient here.
+                    let mingw_directories = path_var
+                        .split(';')
+                        .filter(|dir| dir.contains("mingw"))
+                        .join(";");
+
+                    if !mingw_directories.is_empty() {
+                        cmd.env("PATH", mingw_directories);
+                    }
+                }
+            }
         }
 
         cmd
@@ -357,3 +360,13 @@ pub fn escaped_fake_diff_editor_path() -> String {
     // in it
     diff_editor_path.to_str().unwrap().replace('\\', r"\\")
 }
+
+/// Returns a string with the last line removed.
+///
+/// Use this to remove the root error message containing platform-specific
+/// content for example.
+pub fn strip_last_line(s: &str) -> &str {
+    s.trim_end_matches('\n')
+        .rsplit_once('\n')
+        .map_or(s, |(h, _)| &s[..h.len() + 1])
+}
