@@ -376,7 +376,7 @@ fn test_git_colocated_branch_forget() {
     ◉  230dd059e1b059aefc0da06a2e5a7dbf22362f22 HEAD@git
     ◉  0000000000000000000000000000000000000000
     "###);
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     foo: rlvkpnrz 65b6b74e (empty) (no description set)
       @git: rlvkpnrz 65b6b74e (empty) (no description set)
@@ -387,7 +387,7 @@ fn test_git_colocated_branch_forget() {
     insta::assert_snapshot!(stderr, @"");
     // A forgotten branch is deleted in the git repo. For a detailed demo explaining
     // this, see `test_branch_forget_export` in `test_branch_command.rs`.
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @"");
 }
 
@@ -402,7 +402,7 @@ fn test_git_colocated_conflicting_git_refs() {
     insta::assert_snapshot!(stdout, @"");
     insta::with_settings!({filters => vec![(": The lock for resource.*", ": ...")]}, {
         insta::assert_snapshot!(stderr, @r###"
-        Failed to export some branches:
+        Warning: Failed to export some branches:
           main/sub: Failed to set: A lock could not be obtained for reference "refs/heads/main/sub": ...
         Hint: Git doesn't allow a branch name that looks like a parent directory of
         another (e.g. `foo` and `foo/bar`). Try to rename the branches that failed to
@@ -452,7 +452,7 @@ fn test_git_colocated_checkout_non_empty_working_copy() {
     std::fs::write(workspace_root.join("two"), "y").unwrap();
 
     test_env.jj_cmd_ok(&workspace_root, &["describe", "-m", "two"]);
-    test_env.jj_cmd_ok(&workspace_root, &["co", "@-"]);
+    test_env.jj_cmd_ok(&workspace_root, &["new", "@-"]);
     let (_, stderr) = test_env.jj_cmd_ok(&workspace_root, &["describe", "-m", "new"]);
     insta::assert_snapshot!(stderr, @r###"
     Working copy now at: kkmpptxz 4c049607 (empty) new
@@ -507,6 +507,8 @@ fn test_git_colocated_fetch_deleted_or_moved_branch() {
     let (stdout, stderr) = test_env.jj_cmd_ok(&clone_path, &["git", "fetch"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
+    branch: B_to_delete@origin [deleted] untracked
+    branch: C_to_move@origin   [updated] tracked
     Abandoned 2 commits that are no longer reachable.
     "###);
     // "original C" and "B_to_delete" are abandoned, as the corresponding branches
@@ -553,7 +555,7 @@ fn test_git_colocated_rebase_dirty_working_copy() {
       Use `jj branch list` to see details. Use `jj branch set <name> -r <rev>` to resolve.
     "###);
     insta::assert_snapshot!(stderr, @r###"
-    Failed to export some branches:
+    Warning: Failed to export some branches:
       feature: Modified ref had been deleted in Git
     Done importing changes from the underlying Git repo.
     "###);
