@@ -50,7 +50,7 @@ fn set_up() -> (TestEnvironment, PathBuf) {
 fn test_git_push_nothing() {
     let (test_env, workspace_root) = set_up();
     // Show the setup. `insta` has trouble if this is done inside `set_up()`
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch1: lzmmnrxq 45a3aa29 (empty) description 1
       @origin: lzmmnrxq 45a3aa29 (empty) description 1
@@ -75,12 +75,12 @@ fn test_git_push_current_branch() {
         &workspace_root,
         &["describe", "branch1", "-m", "modified branch1 commit"],
     );
-    test_env.jj_cmd_ok(&workspace_root, &["co", "branch2"]);
+    test_env.jj_cmd_ok(&workspace_root, &["new", "branch2"]);
     test_env.jj_cmd_ok(&workspace_root, &["branch", "set", "branch2"]);
     test_env.jj_cmd_ok(&workspace_root, &["branch", "create", "my-branch"]);
     test_env.jj_cmd_ok(&workspace_root, &["describe", "-m", "foo"]);
     // Check the setup
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch1: lzmmnrxq 19e00bf6 (empty) modified branch1 commit
       @origin (ahead by 1 commits, behind by 1 commits): lzmmnrxq hidden 45a3aa29 (empty) description 1
@@ -104,7 +104,7 @@ fn test_git_push_current_branch() {
       Move branch branch2 from 8476341eb395 to 10ee3363b259
       Add branch my-branch to 10ee3363b259
     "###);
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch1: lzmmnrxq 19e00bf6 (empty) modified branch1 commit
       @origin (ahead by 1 commits, behind by 1 commits): lzmmnrxq hidden 45a3aa29 (empty) description 1
@@ -141,7 +141,7 @@ fn test_git_push_no_matching_branch() {
     let (stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    No branches found in the default push revset, `remote_branches(remote=origin)..@`.
+    Warning: No branches found in the default push revset: remote_branches(remote=origin)..@
     Nothing changed.
     "###);
 }
@@ -149,11 +149,11 @@ fn test_git_push_no_matching_branch() {
 #[test]
 fn test_git_push_matching_branch_unchanged() {
     let (test_env, workspace_root) = set_up();
-    test_env.jj_cmd_ok(&workspace_root, &["co", "branch1"]);
+    test_env.jj_cmd_ok(&workspace_root, &["new", "branch1"]);
     let (stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    No branches found in the default push revset, `remote_branches(remote=origin)..@`.
+    Warning: No branches found in the default push revset: remote_branches(remote=origin)..@
     Nothing changed.
     "###);
 }
@@ -196,7 +196,7 @@ fn test_git_push_other_remote_has_branch() {
     let (stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    No branches found in the default push revset, `remote_branches(remote=origin)..@`.
+    Warning: No branches found in the default push revset: remote_branches(remote=origin)..@
     Nothing changed.
     "###);
     // But it will still get pushed to another remote
@@ -256,7 +256,7 @@ fn test_git_push_locally_created_and_rewritten() {
     // Rewrite it and push again, which would fail if the pushed branch weren't
     // set to "tracking"
     test_env.jj_cmd_ok(&workspace_root, &["describe", "-mlocal 2"]);
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch1: lzmmnrxq 45a3aa29 (empty) description 1
       @origin: lzmmnrxq 45a3aa29 (empty) description 1
@@ -283,7 +283,7 @@ fn test_git_push_multiple() {
     test_env.jj_cmd_ok(&workspace_root, &["branch", "create", "my-branch"]);
     test_env.jj_cmd_ok(&workspace_root, &["describe", "-m", "foo"]);
     // Check the setup
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch1 (deleted)
       @origin: lzmmnrxq 45a3aa29 (empty) description 1
@@ -369,7 +369,7 @@ fn test_git_push_multiple() {
       Force branch branch2 from 8476341eb395 to 15dcdaa4f12f
       Add branch my-branch to 15dcdaa4f12f
     "###);
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch2: yqosqzyt 15dcdaa4 (empty) foo
       @origin: yqosqzyt 15dcdaa4 (empty) foo
@@ -378,11 +378,11 @@ fn test_git_push_multiple() {
     "###);
     let stdout = test_env.jj_cmd_success(&workspace_root, &["log", "-rall()"]);
     insta::assert_snapshot!(stdout, @r###"
-    @  yqosqzyt test.user@example.com 2001-02-03 04:05:17.000 +07:00 branch2 my-branch 15dcdaa4
+    @  yqosqzyt test.user@example.com 2001-02-03 08:05:17 branch2 my-branch 15dcdaa4
     │  (empty) foo
-    │ ◉  rlzusymt test.user@example.com 2001-02-03 04:05:10.000 +07:00 8476341e
+    │ ◉  rlzusymt test.user@example.com 2001-02-03 08:05:10 8476341e
     ├─╯  (empty) description 2
-    │ ◉  lzmmnrxq test.user@example.com 2001-02-03 04:05:08.000 +07:00 45a3aa29
+    │ ◉  lzmmnrxq test.user@example.com 2001-02-03 08:05:08 45a3aa29
     ├─╯  (empty) description 1
     ◉  zzzzzzzz root() 00000000
     "###);
@@ -421,6 +421,55 @@ fn test_git_push_changes() {
     Branch changes to push to origin:
       Force branch push-yostqsxwqrlt from 48d8c7948133 to b5f030322b1d
     "###);
+
+    // specifying the same branch with --change/--branch doesn't break things
+    std::fs::write(workspace_root.join("file"), "modified4").unwrap();
+    let (stdout, stderr) = test_env.jj_cmd_ok(
+        &workspace_root,
+        &["git", "push", "-c=@", "-b=push-yostqsxwqrlt"],
+    );
+    insta::assert_snapshot!(stdout, @"");
+    insta::assert_snapshot!(stderr, @r###"
+    Branch changes to push to origin:
+      Force branch push-yostqsxwqrlt from b5f030322b1d to 4df62cec2ee4
+    "###);
+
+    // try again with --change that moves the branch forward
+    std::fs::write(workspace_root.join("file"), "modified5").unwrap();
+    test_env.jj_cmd_ok(
+        &workspace_root,
+        &[
+            "branch",
+            "set",
+            "-r=@-",
+            "--allow-backwards",
+            "push-yostqsxwqrlt",
+        ],
+    );
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["status"]);
+    insta::assert_snapshot!(stdout, @r###"
+    Working copy changes:
+    M file
+    Working copy : yostqsxw 3e2ce808 bar
+    Parent commit: yqosqzyt fa16a141 push-yostqsxwqrlt* push-yqosqzytrlsw | foo
+    "###);
+    let (stdout, stderr) = test_env.jj_cmd_ok(
+        &workspace_root,
+        &["git", "push", "-c=@", "-b=push-yostqsxwqrlt"],
+    );
+    insta::assert_snapshot!(stdout, @"");
+    insta::assert_snapshot!(stderr, @r###"
+    Branch changes to push to origin:
+      Force branch push-yostqsxwqrlt from 4df62cec2ee4 to 3e2ce808759b
+    "###);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["status"]);
+    insta::assert_snapshot!(stdout, @r###"
+    Working copy changes:
+    M file
+    Working copy : yostqsxw 3e2ce808 push-yostqsxwqrlt | bar
+    Parent commit: yqosqzyt fa16a141 push-yqosqzytrlsw | foo
+    "###);
+
     // Test changing `git.push-branch-prefix`. It causes us to push again.
     let (stdout, stderr) = test_env.jj_cmd_ok(
         &workspace_root,
@@ -436,7 +485,7 @@ fn test_git_push_changes() {
     insta::assert_snapshot!(stderr, @r###"
     Creating branch test-yostqsxwqrlt for revision @
     Branch changes to push to origin:
-      Add branch test-yostqsxwqrlt to b5f030322b1d
+      Add branch test-yostqsxwqrlt to 3e2ce808759b
     "###);
 }
 
@@ -454,15 +503,16 @@ fn test_git_push_revisions() {
     std::fs::write(workspace_root.join("file"), "modified again").unwrap();
 
     // Push an empty set
-    let stderr = test_env.jj_cmd_failure(&workspace_root, &["git", "push", "-r=none()"]);
+    let (_stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push", "-r=none()"]);
     insta::assert_snapshot!(stderr, @r###"
-    Error: Empty revision set
+    Warning: No branches point to the specified revisions: none()
+    Nothing changed.
     "###);
     // Push a revision with no branches
     let (stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push", "-r=@--"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    No branches point to the specified revisions.
+    Warning: No branches point to the specified revisions: @--
     Nothing changed.
     "###);
     // Push a revision with a single branch
@@ -481,6 +531,7 @@ fn test_git_push_revisions() {
     );
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
+    Warning: No branches point to the specified revisions: @--
     Branch changes to push to origin:
       Add branch branch-1 to 7decc7932d9c
     Dry-run requested, not pushing.
@@ -529,8 +580,8 @@ fn test_git_push_mixed() {
     insta::assert_snapshot!(stderr, @r###"
     Creating branch push-yqosqzytrlsw for revision @--
     Branch changes to push to origin:
-      Add branch branch-1 to 7decc7932d9c
       Add branch push-yqosqzytrlsw to fa16a14170fb
+      Add branch branch-1 to 7decc7932d9c
       Add branch branch-2a to 1b45449e18d0
       Add branch branch-2b to 1b45449e18d0
     "###);
@@ -577,7 +628,7 @@ fn test_git_push_conflict() {
     test_env.jj_cmd_ok(&workspace_root, &["describe", "-m", "third"]);
     let stderr = test_env.jj_cmd_failure(&workspace_root, &["git", "push", "--all"]);
     insta::assert_snapshot!(stderr, @r###"
-    Error: Won't push commit 1973d389875c since it has conflicts
+    Error: Won't push commit d9ca3146ade7 since it has conflicts
     "###);
 }
 
@@ -670,11 +721,11 @@ fn test_git_push_deleted() {
     "###);
     let stdout = test_env.jj_cmd_success(&workspace_root, &["log", "-rall()"]);
     insta::assert_snapshot!(stdout, @r###"
-    ◉  rlzusymt test.user@example.com 2001-02-03 04:05:10.000 +07:00 branch2 8476341e
+    ◉  rlzusymt test.user@example.com 2001-02-03 08:05:10 branch2 8476341e
     │  (empty) description 2
-    │ ◉  lzmmnrxq test.user@example.com 2001-02-03 04:05:08.000 +07:00 45a3aa29
+    │ ◉  lzmmnrxq test.user@example.com 2001-02-03 08:05:08 45a3aa29
     ├─╯  (empty) description 1
-    │ @  yqosqzyt test.user@example.com 2001-02-03 04:05:13.000 +07:00 5b36783c
+    │ @  yqosqzyt test.user@example.com 2001-02-03 08:05:13 5b36783c
     ├─╯  (empty) (no description set)
     ◉  zzzzzzzz root() 00000000
     "###);
@@ -706,7 +757,7 @@ fn test_git_push_conflicting_branches() {
     test_env.jj_cmd_ok(&workspace_root, &["branch", "create", "branch2"]);
     test_env.jj_cmd_ok(&workspace_root, &["git", "fetch"]);
     insta::assert_snapshot!(
-        test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]), @r###"
+        test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]), @r###"
     branch1: lzmmnrxq 45a3aa29 (empty) description 1
       @origin: lzmmnrxq 45a3aa29 (empty) description 1
     branch2 (conflicted):
@@ -724,7 +775,7 @@ fn test_git_push_conflicting_branches() {
     let (stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    Branch branch2 is conflicted
+    Warning: Branch branch2 is conflicted
     Hint: Run `jj branch list` to inspect, and use `jj branch set` to fix it up.
     Nothing changed.
     "###);
@@ -741,7 +792,7 @@ fn test_git_push_conflicting_branches() {
     let (stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push", "--all"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    Branch branch2 is conflicted
+    Warning: Branch branch2 is conflicted
     Hint: Run `jj branch list` to inspect, and use `jj branch set` to fix it up.
     Branch changes to push to origin:
       Move branch branch1 from 45a3aa29e907 to fd1d63e031ea
@@ -752,7 +803,7 @@ fn test_git_push_conflicting_branches() {
     let (stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push", "-rall()"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    Branch branch2 is conflicted
+    Warning: Branch branch2 is conflicted
     Hint: Run `jj branch list` to inspect, and use `jj branch set` to fix it up.
     Branch changes to push to origin:
       Move branch branch1 from fd1d63e031ea to 8263cf992d33
@@ -786,7 +837,7 @@ fn test_git_push_tracked_vs_all() {
     test_env.jj_cmd_ok(&workspace_root, &["branch", "delete", "branch2"]);
     test_env.jj_cmd_ok(&workspace_root, &["branch", "untrack", "branch1@origin"]);
     test_env.jj_cmd_ok(&workspace_root, &["branch", "create", "branch3"]);
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch1: vruxwmqv a25f24af (empty) moved branch1
     branch1@origin: lzmmnrxq 45a3aa29 (empty) description 1
@@ -808,7 +859,7 @@ fn test_git_push_tracked_vs_all() {
 
     // Untrack the last remaining tracked branch.
     test_env.jj_cmd_ok(&workspace_root, &["branch", "untrack", "branch2@origin"]);
-    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all"]);
+    let stdout = test_env.jj_cmd_success(&workspace_root, &["branch", "list", "--all-remotes"]);
     insta::assert_snapshot!(stdout, @r###"
     branch1: vruxwmqv a25f24af (empty) moved branch1
     branch1@origin: lzmmnrxq 45a3aa29 (empty) description 1
@@ -840,7 +891,7 @@ fn test_git_push_tracked_vs_all() {
     //   instead of showing an error here.
     let (_stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push", "--all"]);
     insta::assert_snapshot!(stderr, @r###"
-    Non-tracking remote branch branch1@origin exists
+    Warning: Non-tracking remote branch branch1@origin exists
     Hint: Run `jj branch track branch1@origin` to import the remote branch.
     Branch changes to push to origin:
       Add branch branch3 to 998d6a7853d9
@@ -856,7 +907,7 @@ fn test_git_push_moved_forward_untracked() {
     test_env.jj_cmd_ok(&workspace_root, &["branch", "untrack", "branch1@origin"]);
     let (_stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push"]);
     insta::assert_snapshot!(stderr, @r###"
-    Non-tracking remote branch branch1@origin exists
+    Warning: Non-tracking remote branch branch1@origin exists
     Hint: Run `jj branch track branch1@origin` to import the remote branch.
     Nothing changed.
     "###);
@@ -874,7 +925,7 @@ fn test_git_push_moved_sideways_untracked() {
     test_env.jj_cmd_ok(&workspace_root, &["branch", "untrack", "branch1@origin"]);
     let (_stdout, stderr) = test_env.jj_cmd_ok(&workspace_root, &["git", "push"]);
     insta::assert_snapshot!(stderr, @r###"
-    Non-tracking remote branch branch1@origin exists
+    Warning: Non-tracking remote branch branch1@origin exists
     Hint: Run `jj branch track branch1@origin` to import the remote branch.
     Nothing changed.
     "###);
