@@ -29,49 +29,28 @@ fn test_log_with_empty_revision() {
 }
 
 #[test]
-fn test_log_legacy_range_operator() {
+fn test_log_with_no_template() {
     let test_env = TestEnvironment::default();
     test_env.jj_cmd_ok(test_env.env_root(), &["init", "repo", "--git"]);
     let repo_path = test_env.env_root().join("repo");
 
-    let (stdout, stderr) = test_env.jj_cmd_ok(&repo_path, &["log", "-r=@:"]);
-    insta::assert_snapshot!(stdout, @r###"
-    @  qpvuntsm test.user@example.com 2001-02-03 04:05:07.000 +07:00 230dd059
-    │  (empty) (no description set)
-    ~
-    "###);
+    let stderr = test_env.jj_cmd_cli_error(&repo_path, &["log", "-T"]);
     insta::assert_snapshot!(stderr, @r###"
-    The `:` revset operator is deprecated. Please switch to `::`.
-    "###);
-    let (stdout, stderr) = test_env.jj_cmd_ok(&repo_path, &["log", "-r=:@"]);
-    insta::assert_snapshot!(stdout, @r###"
-    @  qpvuntsm test.user@example.com 2001-02-03 04:05:07.000 +07:00 230dd059
-    │  (empty) (no description set)
-    ◉  zzzzzzzz root() 00000000
-    "###);
-    insta::assert_snapshot!(stderr, @r###"
-    The `:` revset operator is deprecated. Please switch to `::`.
-    "###);
-    let (stdout, stderr) = test_env.jj_cmd_ok(&repo_path, &["log", "-r=root():@"]);
-    insta::assert_snapshot!(stdout, @r###"
-    @  qpvuntsm test.user@example.com 2001-02-03 04:05:07.000 +07:00 230dd059
-    │  (empty) (no description set)
-    ◉  zzzzzzzz root() 00000000
-    "###);
-    insta::assert_snapshot!(stderr, @r###"
-    The `:` revset operator is deprecated. Please switch to `::`.
-    "###);
-    let (stdout, stderr) = test_env.jj_cmd_ok(
-        &repo_path,
-        &["log", "-r=x", "--config-toml", "revset-aliases.x = '@:'"],
-    );
-    insta::assert_snapshot!(stdout, @r###"
-    @  qpvuntsm test.user@example.com 2001-02-03 04:05:07.000 +07:00 230dd059
-    │  (empty) (no description set)
-    ~
-    "###);
-    insta::assert_snapshot!(stderr, @r###"
-    The `:` revset operator is deprecated. Please switch to `::`.
+    error: a value is required for '--template <TEMPLATE>' but none was supplied
+
+    For more information, try '--help'.
+    Hint: The following template aliases are defined:
+    - builtin_change_id_with_hidden_and_divergent_info
+    - builtin_log_comfortable
+    - builtin_log_compact
+    - builtin_log_detailed
+    - builtin_log_oneline
+    - builtin_op_log_comfortable
+    - builtin_op_log_compact
+    - commit_summary_separator
+    - description_placeholder
+    - email_placeholder
+    - name_placeholder
     "###);
 }
 
@@ -420,7 +399,7 @@ fn test_log_shortest_accessors() {
     "###);
 
     // Can get shorter prefixes in configured revset
-    test_env.add_config(r#"revsets.short-prefixes = "(@----):""#);
+    test_env.add_config(r#"revsets.short-prefixes = "(@----)::""#);
     insta::assert_snapshot!(
         render("::@", r#"format_id(change_id) ++ " " ++ format_id(commit_id) ++ "\n""#),
         @r###"
@@ -466,7 +445,8 @@ fn test_log_bad_short_prefixes() {
     let stderr = test_env.jj_cmd_failure(&repo_path, &["status"]);
     insta::assert_snapshot!(stderr,
         @r###"
-    Config error: Invalid `revsets.short-prefixes`:  --> 1:1
+    Config error: Invalid `revsets.short-prefixes`
+    Caused by:  --> 1:1
       |
     1 | !nval!d
       | ^---
@@ -653,10 +633,10 @@ fn test_log_prefix_highlight_counts_hidden_commits() {
     insta::assert_snapshot!(
         test_env.jj_cmd_success(&repo_path, &["log", "-T", prefix_format]),
         @r###"
-    @  Change w[qnwkozpkust] 44[4c3c5066d3]
-    │ ◉  Change q[pvuntsmwlqt] initial ba[1a30916d29] original
+    @  Change wq[nwkozpkust] 44[4c3c5066d3]
+    │ ◉  Change qpv[untsmwlqt] initial ba[1a30916d29] original
     ├─╯
-    ◉  Change z[zzzzzzzzzzz] 00[0000000000]
+    ◉  Change zzz[zzzzzzzzz] 00[0000000000]
     "###
     );
     insta::assert_snapshot!(
@@ -668,7 +648,7 @@ fn test_log_prefix_highlight_counts_hidden_commits() {
     insta::assert_snapshot!(
         test_env.jj_cmd_success(&repo_path, &["log", "-r", "44", "-T", prefix_format]),
         @r###"
-    @  Change w[qnwkozpkust] 44[4c3c5066d3]
+    @  Change wq[nwkozpkust] 44[4c3c5066d3]
     │
     ~
     "###
@@ -694,8 +674,8 @@ fn test_log_short_shortest_length_parameter() {
     "###);
     insta::assert_snapshot!(
         render(r#"commit_id.short(-100) ++ "|" ++ commit_id.shortest(-100)"#), @r###"
-    @  |2
-    ◉  |0
+    @  <Error: out of range integral type conversion attempted>|<Error: out of range integral type conversion attempted>
+    ◉  <Error: out of range integral type conversion attempted>|<Error: out of range integral type conversion attempted>
     "###);
     insta::assert_snapshot!(
         render(r#"commit_id.short(100) ++ "|" ++ commit_id.shortest(100)"#), @r###"
@@ -713,7 +693,7 @@ fn test_log_author_format() {
     insta::assert_snapshot!(
         test_env.jj_cmd_success(&repo_path, &["log", "--revisions=@"]),
         @r###"
-    @  qpvuntsm test.user@example.com 2001-02-03 04:05:07.000 +07:00 230dd059
+    @  qpvuntsm test.user@example.com 2001-02-03 08:05:07 230dd059
     │  (empty) (no description set)
     ~
     "###
@@ -731,7 +711,7 @@ fn test_log_author_format() {
             ],
         ),
         @r###"
-    @  qpvuntsm test.user 2001-02-03 04:05:07.000 +07:00 230dd059
+    @  qpvuntsm test.user 2001-02-03 08:05:07 230dd059
     │  (empty) (no description set)
     ~
     "###
@@ -968,7 +948,7 @@ fn test_log_warn_path_might_be_revset() {
     ~
     "###);
     insta::assert_snapshot!(stderr, @r###"
-    warning: The argument "." is being interpreted as a path, but this is often not useful because all non-empty commits touch '.'.  If you meant to show the working copy commit, pass -r '@' instead.
+    Warning: The argument "." is being interpreted as a path, but this is often not useful because all non-empty commits touch '.'.  If you meant to show the working copy commit, pass -r '@' instead.
     "###);
 
     // ...but checking `jj log .` makes sense in a subdirectory.
@@ -982,14 +962,14 @@ fn test_log_warn_path_might_be_revset() {
     let (stdout, stderr) = test_env.jj_cmd_ok(&repo_path, &["log", "@", "-T", "description"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    warning: The argument "@" is being interpreted as a path. To specify a revset, pass -r "@" instead.
+    Warning: The argument "@" is being interpreted as a path. To specify a revset, pass -r "@" instead.
     "###);
 
     // Warn when there's no path with the provided name.
     let (stdout, stderr) = test_env.jj_cmd_ok(&repo_path, &["log", "file2", "-T", "description"]);
     insta::assert_snapshot!(stdout, @"");
     insta::assert_snapshot!(stderr, @r###"
-    warning: The argument "file2" is being interpreted as a path. To specify a revset, pass -r "file2" instead.
+    Warning: The argument "file2" is being interpreted as a path. To specify a revset, pass -r "file2" instead.
     "###);
 
     // If an explicit revision is provided, then suppress the warning.
@@ -1021,6 +1001,15 @@ fn test_default_revset() {
             .lines()
             .count()
     );
+
+    // The default revset is not used if a path is specified
+    insta::assert_snapshot!(
+        test_env.jj_cmd_success(&repo_path, &["log", "file1", "-T", "description"]),
+        @r###"
+    @  add a file
+    │
+    ~
+    "###);
 }
 
 #[test]
@@ -1147,7 +1136,7 @@ fn test_graph_styles() {
         &["new", "-m", "merge", r#"description("main branch 1")"#, "@"],
     );
 
-    // Default (legacy) style
+    // Default (curved) style
     let stdout = test_env.jj_cmd_success(&repo_path, &["log", "-T=description"]);
     insta::assert_snapshot!(stdout, @r###"
     @    merge
@@ -1260,32 +1249,30 @@ fn test_log_word_wrap() {
 
     // ui.log-word-wrap option applies to both graph/no-graph outputs
     insta::assert_snapshot!(render(&["log", "-r@"], 40, false), @r###"
-    @  mzvwutvl test.user@example.com 2001-02-03 04:05:11.000 +07:00 68518a7e
+    @  mzvwutvl test.user@example.com 2001-02-03 08:05:11 68518a7e
     │  (empty) merge
     ~
     "###);
     insta::assert_snapshot!(render(&["log", "-r@"], 40, true), @r###"
     @  mzvwutvl test.user@example.com
-    │  2001-02-03 04:05:11.000 +07:00
-    ~  68518a7e
-       (empty) merge
+    │  2001-02-03 08:05:11 68518a7e
+    ~  (empty) merge
     "###);
     insta::assert_snapshot!(render(&["log", "--no-graph", "-r@"], 40, false), @r###"
-    mzvwutvl test.user@example.com 2001-02-03 04:05:11.000 +07:00 68518a7e
+    mzvwutvl test.user@example.com 2001-02-03 08:05:11 68518a7e
     (empty) merge
     "###);
     insta::assert_snapshot!(render(&["log", "--no-graph", "-r@"], 40, true), @r###"
     mzvwutvl test.user@example.com
-    2001-02-03 04:05:11.000 +07:00 68518a7e
+    2001-02-03 08:05:11 68518a7e
     (empty) merge
     "###);
 
     // Color labels should be preserved
     insta::assert_snapshot!(render(&["log", "-r@", "--color=always"], 40, true), @r###"
     @  [1m[38;5;13mm[38;5;8mzvwutvl[39m [38;5;3mtest.user@example.com[39m[0m
-    │  [1m[38;5;14m2001-02-03 04:05:11.000 +07:00[39m[0m
-    ~  [1m[38;5;12m6[38;5;8m8518a7e[39m[0m
-       [1m[38;5;10m(empty)[39m merge[0m
+    │  [1m[38;5;14m2001-02-03 08:05:11[39m [38;5;12m6[38;5;8m8518a7e[39m[0m
+    ~  [1m[38;5;10m(empty)[39m merge[0m
     "###);
 
     // Graph width should be subtracted from the term width
@@ -1310,36 +1297,13 @@ fn test_log_word_wrap() {
        4 5 6 7
        8 9
     "###);
-    insta::assert_snapshot!(
-        render(&["log", "-T", template, "--config-toml=ui.graph.style='legacy'"], 9, true),
-        @r###"
-    @   0 1 2
-    |\  3 4 5
-    | | 6 7 8
-    | | 9
-    | o 0 1 2
-    | | 3 4 5
-    | | 6 7 8
-    | | 9
-    | o 0 1 2
-    |/  3 4 5
-    |   6 7 8
-    |   9
-    o 0 1 2 3
-    | 4 5 6 7
-    | 8 9
-    o 0 1 2 3
-      4 5 6 7
-      8 9
-    "###);
 
     // Shouldn't panic with $COLUMNS < graph_width
     insta::assert_snapshot!(render(&["log", "-r@"], 0, true), @r###"
     @  mzvwutvl
     │  test.user@example.com
     ~  2001-02-03
-       04:05:11.000
-       +07:00
+       08:05:11
        68518a7e
        (empty)
        merge
@@ -1348,10 +1312,178 @@ fn test_log_word_wrap() {
     @  mzvwutvl
     │  test.user@example.com
     ~  2001-02-03
-       04:05:11.000
-       +07:00
+       08:05:11
        68518a7e
        (empty)
        merge
     "###);
 }
+
+#[test]
+fn test_elided() {
+    // Test that elided commits are shown as synthetic nodes.
+    let test_env = TestEnvironment::default();
+    test_env.jj_cmd_ok(test_env.env_root(), &["init", "repo", "--git"]);
+    let repo_path = test_env.env_root().join("repo");
+
+    test_env.jj_cmd_ok(&repo_path, &["describe", "-m", "initial"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "-m", "main branch 1"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "-m", "main branch 2"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "@--", "-m", "side branch 1"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "-m", "side branch 2"]);
+    test_env.jj_cmd_ok(
+        &repo_path,
+        &["new", "-m", "merge", r#"description("main branch 2")"#, "@"],
+    );
+
+    let get_log = |revs: &str| -> String {
+        test_env.jj_cmd_success(
+            &repo_path,
+            &["log", "-T", r#"description ++ "\n""#, "-r", revs],
+        )
+    };
+
+    // Test the setup
+    insta::assert_snapshot!(get_log("::"), @r###"
+    @    merge
+    ├─╮
+    │ ◉  side branch 2
+    │ │
+    │ ◉  side branch 1
+    │ │
+    ◉ │  main branch 2
+    │ │
+    ◉ │  main branch 1
+    ├─╯
+    ◉  initial
+    │
+    ◉
+    "###);
+
+    // Elide some commits from each side of the merge. It's unclear that a revision
+    // was skipped on the left side.
+    test_env.add_config("ui.log-synthetic-elided-nodes = false");
+    insta::assert_snapshot!(get_log("@ | @- | description(initial)"), @r###"
+    @    merge
+    ├─╮
+    │ ◉  side branch 2
+    │ ╷
+    ◉ ╷  main branch 2
+    ├─╯
+    ◉  initial
+    │
+    ~
+    "###);
+
+    // Elide shared commits. It's unclear that a revision was skipped on the right
+    // side (#1252).
+    insta::assert_snapshot!(get_log("@-- | root()"), @r###"
+    ◉  side branch 1
+    ╷
+    ╷ ◉  main branch 1
+    ╭─╯
+    ◉
+    "###);
+
+    // Now test the same thing with synthetic nodes for elided commits
+
+    // Elide some commits from each side of the merge
+    test_env.add_config("ui.log-synthetic-elided-nodes = true");
+    insta::assert_snapshot!(get_log("@ | @- | description(initial)"), @r###"
+    @    merge
+    ├─╮
+    │ ◉  side branch 2
+    │ │
+    │ ◌  (elided revisions)
+    ◉ │  main branch 2
+    │ │
+    ◌ │  (elided revisions)
+    ├─╯
+    ◉  initial
+    │
+    ~
+    "###);
+
+    // Elide shared commits. To keep the implementation simple, it still gets
+    // rendered as two synthetic nodes.
+    insta::assert_snapshot!(get_log("@-- | root()"), @r###"
+    ◉  side branch 1
+    │
+    ◌  (elided revisions)
+    │ ◉  main branch 1
+    │ │
+    │ ◌  (elided revisions)
+    ├─╯
+    ◉
+    "###);
+}
+
+#[test]
+fn test_log_with_custom_symbols() {
+    // Test that elided commits are shown as synthetic nodes.
+    let test_env = TestEnvironment::default();
+    test_env.jj_cmd_ok(test_env.env_root(), &["init", "repo", "--git"]);
+    let repo_path = test_env.env_root().join("repo");
+
+    test_env.jj_cmd_ok(&repo_path, &["describe", "-m", "initial"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "-m", "main branch 1"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "-m", "main branch 2"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "@--", "-m", "side branch 1"]);
+    test_env.jj_cmd_ok(&repo_path, &["new", "-m", "side branch 2"]);
+    test_env.jj_cmd_ok(
+        &repo_path,
+        &["new", "-m", "merge", r#"description("main branch 2")"#, "@"],
+    );
+
+    let get_log = |revs: &str| -> String {
+        test_env.jj_cmd_success(
+            &repo_path,
+            &["log", "-T", r#"description ++ "\n""#, "-r", revs],
+        )
+    };
+
+    // Simple test with showing default and elided nodes.
+    test_env.add_config(
+        r###"
+        ui.log-synthetic-elided-nodes = true
+        templates.log_node = 'if(self, if(current_working_copy, "$", if(root, "┴", "┝")), "🮀")'
+        "###,
+    );
+    insta::assert_snapshot!(get_log("@ | @- | description(initial) | root()"), @r###"
+    $    merge
+    ├─╮
+    │ ┝  side branch 2
+    │ │
+    │ 🮀  (elided revisions)
+    ┝ │  main branch 2
+    │ │
+    🮀 │  (elided revisions)
+    ├─╯
+    ┝  initial
+    │
+    ┴
+    "###);
+
+    // Simple test with showing default and elided nodes, ascii style.
+    test_env.add_config(
+        r###"
+        ui.log-synthetic-elided-nodes = true
+        ui.graph.style = 'ascii'
+        templates.log_node = 'if(self, if(current_working_copy, "$", if(root, "^", "*")), ":")'
+        "###,
+    );
+    insta::assert_snapshot!(get_log("@ | @- | description(initial) | root()"), @r###"
+    $    merge
+    |\
+    | *  side branch 2
+    | |
+    | :  (elided revisions)
+    * |  main branch 2
+    | |
+    : |  (elided revisions)
+    |/
+    *  initial
+    |
+    ^
+    "###);
+}
