@@ -21,8 +21,6 @@ use std::fs::{File, Metadata, OpenOptions};
 use std::io::{Read, Write};
 use std::ops::Range;
 #[cfg(unix)]
-use std::os::unix::fs::symlink;
-#[cfg(unix)]
 use std::os::unix::fs::PermissionsExt;
 use std::path::{Path, PathBuf};
 use std::sync::mpsc::{channel, Sender};
@@ -46,6 +44,7 @@ use crate::backend::{
 };
 use crate::commit::Commit;
 use crate::conflicts::{self, materialize_tree_value, MaterializedTreeValue};
+use crate::file_util::{check_symlink_support, try_symlink};
 #[cfg(feature = "watchman")]
 use crate::fsmonitor::watchman;
 use crate::fsmonitor::FsmonitorKind;
@@ -197,6 +196,10 @@ impl FileStatesMap {
         .collect();
     }
 
+    fn clear(&mut self) {
+        self.data.clear();
+    }
+
     /// Returns read-only map containing all file states.
     fn all(&self) -> FileStates<'_> {
         FileStates::from_sorted(&self.data)
@@ -289,6 +292,7 @@ pub struct TreeState {
     // Currently only path prefixes
     sparse_patterns: Vec<RepoPathBuf>,
     own_mtime: MillisSinceEpoch,
+    symlink_support: bool,
 
     /// The most recent clock value returned by Watchman. Will only be set if
     /// the repo is configured to use the Watchman filesystem monitor and
@@ -544,6 +548,7 @@ impl TreeState {
             file_states: FileStatesMap::new(),
             sparse_patterns: vec![RepoPathBuf::root()],
             own_mtime: MillisSinceEpoch(0),
+            symlink_support: check_symlink_support().unwrap_or(false),
             watchman_clock: None,
         }
     }
@@ -677,18 +682,29 @@ impl TreeState {
         path: &RepoPath,
         disk_path: &Path,
     ) -> Result<SymlinkId, SnapshotError> {
-        let target = disk_path.read_link().map_err(|err| SnapshotError::Other {
-            message: format!("Failed to read symlink {}", disk_path.display()),
-            err: err.into(),
-        })?;
-        let str_target =
-            target
-                .to_str()
-                .ok_or_else(|| SnapshotError::InvalidUtf8SymlinkTarget {
+        if self.symlink_support {
+            let target = disk_path.read_link().map_err(|err| SnapshotError::Other {
+                message: format!("Failed to read symlink {}", disk_path.display()),
+                err: err.into(),
+            })?;
+            let str_target =
+                target
+                    .to_str()
+                    .ok_or_else(|| SnapshotError::InvalidUtf8SymlinkTarget {
+                        path: disk_path.to_path_buf(),
+                    })?;
+            Ok(self.store.write_symlink(path, str_target)?)
+        } else {
+            let target = fs::read(disk_path).map_err(|err| SnapshotError::Other {
+                message: format!("Failed to read file {}", disk_path.display()),
+                err: err.into(),
+            })?;
+            let string_target =
+                String::from_utf8(target).map_err(|_| SnapshotError::InvalidUtf8SymlinkTarget {
                     path: disk_path.to_path_buf(),
-                    target: target.clone(),
                 })?;
-        Ok(self.store.write_symlink(path, str_target)?)
+            Ok(self.store.write_symlink(path, &string_target)?)
+        }
     }
 
     fn reset_watchman(&mut self) {
@@ -725,7 +741,7 @@ impl TreeState {
 
         let sparse_matcher = self.sparse_matcher();
 
-        let fsmonitor_clock_needs_save = fsmonitor_kind.is_some();
+        let fsmonitor_clock_needs_save = fsmonitor_kind != FsmonitorKind::None;
         let mut is_dirty = fsmonitor_clock_needs_save;
         let FsmonitorMatcher {
             matcher: fsmonitor_matcher,
@@ -849,8 +865,8 @@ impl TreeState {
             return Ok(());
         }
 
-        let git_ignore =
-            git_ignore.chain_with_file(&dir.to_internal_dir_string(), disk_dir.join(".gitignore"));
+        let git_ignore = git_ignore
+            .chain_with_file(&dir.to_internal_dir_string(), disk_dir.join(".gitignore"))?;
         let dir_entries = disk_dir
             .read_dir()
             .unwrap()
@@ -998,13 +1014,13 @@ impl TreeState {
     #[instrument(skip_all)]
     fn make_fsmonitor_matcher(
         &self,
-        fsmonitor_kind: Option<FsmonitorKind>,
+        fsmonitor_kind: FsmonitorKind,
     ) -> Result<FsmonitorMatcher, SnapshotError> {
         let (watchman_clock, changed_files) = match fsmonitor_kind {
-            None => (None, None),
-            Some(FsmonitorKind::Test { changed_files }) => (None, Some(changed_files)),
+            FsmonitorKind::None => (None, None),
+            FsmonitorKind::Test { changed_files } => (None, Some(changed_files)),
             #[cfg(feature = "watchman")]
-            Some(FsmonitorKind::Watchman) => match self.query_watchman() {
+            FsmonitorKind::Watchman => match self.query_watchman() {
                 Ok((watchman_clock, changed_files)) => (Some(watchman_clock.into()), changed_files),
                 Err(err) => {
                     tracing::warn!(?err, "Failed to query filesystem monitor");
@@ -1012,7 +1028,7 @@ impl TreeState {
                 }
             },
             #[cfg(not(feature = "watchman"))]
-            Some(FsmonitorKind::Watchman) => {
+            FsmonitorKind::Watchman => {
                 return Err(SnapshotError::Other {
                     message: "Failed to query the filesystem monitor".to_string(),
                     err: "Cannot query Watchman because jj was not compiled with the `watchman` \
@@ -1027,7 +1043,7 @@ impl TreeState {
                 let repo_paths = trace_span!("processing fsmonitor paths").in_scope(|| {
                     changed_files
                         .into_iter()
-                        .filter_map(RepoPathBuf::from_relative_path)
+                        .filter_map(|path| RepoPathBuf::from_relative_path(path).ok())
                         .collect_vec()
                 });
 
@@ -1062,14 +1078,31 @@ impl TreeState {
         if clean {
             Ok(None)
         } else {
-            let new_file_type = new_file_state.file_type.clone();
             let current_tree_values = current_tree.path_value(repo_path);
-            let new_tree_values = self.write_path_to_store(
-                repo_path,
-                &disk_path,
-                &current_tree_values,
-                new_file_type,
-            )?;
+            let new_file_type = if !self.symlink_support {
+                let mut new_file_type = new_file_state.file_type.clone();
+                if matches!(new_file_type, FileType::Normal { .. })
+                    && matches!(current_tree_values.as_normal(), Some(TreeValue::Symlink(_)))
+                {
+                    new_file_type = FileType::Symlink;
+                }
+                new_file_type
+            } else {
+                new_file_state.file_type.clone()
+            };
+            let new_tree_values = match new_file_type {
+                FileType::Normal { executable } => self.write_path_to_store(
+                    repo_path,
+                    &disk_path,
+                    &current_tree_values,
+                    executable,
+                )?,
+                FileType::Symlink => {
+                    let id = self.write_symlink_to_store(repo_path, &disk_path)?;
+                    Merge::normal(TreeValue::Symlink(id))
+                }
+                FileType::GitSubmodule => panic!("git submodule cannot be written to store"),
+            };
             if new_tree_values != current_tree_values {
                 Ok(Some(new_tree_values))
             } else {
@@ -1083,17 +1116,8 @@ impl TreeState {
         repo_path: &RepoPath,
         disk_path: &Path,
         current_tree_values: &MergedTreeValue,
-        file_type: FileType,
+        executable: FileExecutableFlag,
     ) -> Result<MergedTreeValue, SnapshotError> {
-        let executable = match file_type {
-            FileType::Normal { executable } => executable,
-            FileType::Symlink => {
-                let id = self.write_symlink_to_store(repo_path, disk_path)?;
-                return Ok(Merge::normal(TreeValue::Symlink(id)));
-            }
-            FileType::GitSubmodule => panic!("git submodule cannot be written to store"),
-        };
-
         // If the file contained a conflict before and is now a normal file on disk, we
         // try to parse any conflict markers in the file into a conflict.
         if let Some(current_tree_value) = current_tree_values.as_resolved() {
@@ -1177,24 +1201,16 @@ impl TreeState {
         Ok(FileState::for_file(executable, size, &metadata))
     }
 
-    #[cfg_attr(windows, allow(unused_variables))]
     fn write_symlink(&self, disk_path: &Path, target: String) -> Result<FileState, CheckoutError> {
-        #[cfg(windows)]
-        {
-            println!("ignoring symlink at {}", disk_path.display());
-        }
-        #[cfg(unix)]
-        {
-            let target = PathBuf::from(&target);
-            symlink(&target, disk_path).map_err(|err| CheckoutError::Other {
-                message: format!(
-                    "Failed to create symlink from {} to {}",
-                    disk_path.display(),
-                    target.display()
-                ),
-                err: err.into(),
-            })?;
-        }
+        let target = PathBuf::from(&target);
+        try_symlink(&target, disk_path).map_err(|err| CheckoutError::Other {
+            message: format!(
+                "Failed to create symlink from {} to {}",
+                disk_path.display(),
+                target.display()
+            ),
+            err: err.into(),
+        })?;
         let metadata = disk_path
             .symlink_metadata()
             .map_err(|err| checkout_error_for_stat_error(err, disk_path))?;
@@ -1361,7 +1377,11 @@ impl TreeState {
                     ..
                 } => self.write_file(&disk_path, &mut reader, executable)?,
                 MaterializedTreeValue::Symlink { id: _, target } => {
-                    self.write_symlink(&disk_path, target)?
+                    if self.symlink_support {
+                        self.write_symlink(&disk_path, target)?
+                    } else {
+                        self.write_file(&disk_path, &mut target.as_bytes(), false)?
+                    }
                 }
                 MaterializedTreeValue::GitSubmodule(_) => {
                     println!("ignoring git submodule at {path:?}");
@@ -1436,6 +1456,12 @@ impl TreeState {
         self.tree_id = new_tree.id();
         Ok(())
     }
+
+    pub async fn recover(&mut self, new_tree: &MergedTree) -> Result<(), ResetError> {
+        self.file_states.clear();
+        self.tree_id = self.store.empty_merged_tree_id();
+        self.reset(new_tree).await
+    }
 }
 
 fn checkout_error_for_stat_error(err: std::io::Error, path: &Path) -> CheckoutError {
@@ -1753,6 +1779,20 @@ impl LockedWorkingCopy for LockedLocalWorkingCopy {
         Ok(())
     }
 
+    fn recover(&mut self, commit: &Commit) -> Result<(), ResetError> {
+        let new_tree = commit.tree()?;
+        self.wc
+            .tree_state_mut()
+            .map_err(|err| ResetError::Other {
+                message: "Failed to read the working copy state".to_string(),
+                err: err.into(),
+            })?
+            .recover(&new_tree)
+            .block_on()?;
+        self.tree_state_dirty = true;
+        Ok(())
+    }
+
     fn sparse_patterns(&self) -> Result<&[RepoPathBuf], WorkingCopyStateError> {
         self.wc.sparse_patterns()
     }
