@@ -30,7 +30,7 @@ use tracing::instrument;
 use self::dirty_cell::DirtyCell;
 use crate::backend::{
     Backend, BackendError, BackendInitError, BackendLoadError, BackendResult, ChangeId, CommitId,
-    MergedTreeId, ObjectId, SigningFn,
+    MergedTreeId, SigningFn,
 };
 use crate::commit::{Commit, CommitByCommitterTimestamp};
 use crate::commit_builder::CommitBuilder;
@@ -38,8 +38,9 @@ use crate::default_index::DefaultIndexStore;
 use crate::default_submodule_store::DefaultSubmoduleStore;
 use crate::file_util::{IoResultExt as _, PathError};
 use crate::git_backend::GitBackend;
-use crate::index::{HexPrefix, Index, IndexStore, MutableIndex, PrefixResolution, ReadonlyIndex};
+use crate::index::{ChangeIdIndex, Index, IndexStore, MutableIndex, ReadonlyIndex};
 use crate::local_backend::LocalBackend;
+use crate::object_id::{HexPrefix, ObjectId, PrefixResolution};
 use crate::op_heads_store::{self, OpHeadResolutionError, OpHeadsStore};
 use crate::op_store::{
     OpStore, OpStoreError, OperationId, RefTarget, RemoteRef, RemoteRefState, WorkspaceId,
@@ -48,7 +49,6 @@ use crate::operation::Operation;
 use crate::refs::{
     diff_named_ref_targets, diff_named_remote_refs, merge_ref_targets, merge_remote_refs,
 };
-use crate::revset::{ChangeIdIndex, RevsetExpression};
 use crate::rewrite::{DescendantRebaser, RebaseOptions};
 use crate::settings::{RepoSettings, UserSettings};
 use crate::signing::{SignInitError, Signer};
@@ -57,7 +57,6 @@ use crate::simple_op_store::SimpleOpStore;
 use crate::store::Store;
 use crate::submodule_store::SubmoduleStore;
 use crate::transaction::Transaction;
-use crate::tree::TreeMergeError;
 use crate::view::View;
 use crate::{backend, dag_walk, op_store, revset};
 
@@ -167,29 +166,13 @@ impl ReadonlyRepo {
         fs::write(&op_store_type_path, op_store.name()).context(&op_store_type_path)?;
         let op_store: Arc<dyn OpStore> = Arc::from(op_store);
 
-        let mut root_view = op_store::View::default();
-        root_view.head_ids.insert(store.root_commit_id().clone());
-        root_view
-            .public_head_ids
-            .insert(store.root_commit_id().clone());
-
         let op_heads_path = repo_path.join("op_heads");
         fs::create_dir(&op_heads_path).context(&op_heads_path)?;
-        let operation_metadata =
-            crate::transaction::create_op_metadata(user_settings, "initialize repo".to_string());
-        let root_view_id = op_store.write_view(&root_view).unwrap();
-        let init_operation = op_store::Operation {
-            view_id: root_view_id,
-            parents: vec![],
-            metadata: operation_metadata,
-        };
-        let init_operation_id = op_store.write_operation(&init_operation).unwrap();
-        let init_operation = Operation::new(op_store.clone(), init_operation_id, init_operation);
         let op_heads_store = op_heads_store_initializer(user_settings, &op_heads_path);
-        op_heads_store.update_op_heads(&[], init_operation.id());
         let op_heads_type_path = op_heads_path.join("type");
         fs::write(&op_heads_type_path, op_heads_store.name()).context(&op_heads_type_path)?;
-        let op_heads_store = Arc::from(op_heads_store);
+        op_heads_store.update_op_heads(&[], op_store.root_operation_id());
+        let op_heads_store: Arc<dyn OpHeadsStore> = Arc::from(op_heads_store);
 
         let index_path = repo_path.join("index");
         fs::create_dir(&index_path).context(&index_path)?;
@@ -206,20 +189,33 @@ impl ReadonlyRepo {
             .context(&submodule_store_type_path)?;
         let submodule_store = Arc::from(submodule_store);
 
-        let view = View::new(root_view);
-        Ok(Arc::new(ReadonlyRepo {
+        let root_operation_data = op_store
+            .read_operation(op_store.root_operation_id())
+            .expect("failed to read root operation");
+        let root_operation = Operation::new(
+            op_store.clone(),
+            op_store.root_operation_id().clone(),
+            root_operation_data,
+        );
+        let root_view = root_operation.view().expect("failed to read root view");
+        let repo = Arc::new(ReadonlyRepo {
             repo_path,
             store,
             op_store,
             op_heads_store,
-            operation: init_operation,
+            operation: root_operation,
             settings: repo_settings,
             index_store,
             index: OnceCell::new(),
             change_id_index: OnceCell::new(),
-            view,
+            view: root_view,
             submodule_store,
-        }))
+        });
+        let mut tx = repo.start_transaction(user_settings);
+        tx.mut_repo()
+            .add_head(&repo.store().root_commit())
+            .expect("failed to add root commit as head");
+        Ok(tx.commit("initialize repo"))
     }
 
     pub fn loader(&self) -> RepoLoader {
@@ -265,14 +261,8 @@ impl ReadonlyRepo {
     fn change_id_index(&self) -> &dyn ChangeIdIndex {
         self.change_id_index
             .get_or_init(|| {
-                // TODO: maybe add abstraction over 'static/'index revset
-                // evaluation, and use it.
-                let expression = RevsetExpression::all().resolve_programmatic(self);
-                let revset = self
-                    .readonly_index()
-                    .evaluate_revset_static(&expression, self.store())
-                    .unwrap();
-                revset.change_id_index()
+                self.readonly_index()
+                    .change_id_index(&mut self.view().heads().iter())
             })
             .as_ref()
     }
@@ -414,7 +404,7 @@ pub enum StoreLoadError {
         store: &'static str,
         store_type: String,
     },
-    #[error("Failed to read {store} backend type: {source}")]
+    #[error("Failed to read {store} backend type")]
     ReadError {
         store: &'static str,
         source: PathError,
@@ -589,7 +579,7 @@ pub fn read_store_type_compat(
 #[derive(Debug, Error)]
 pub enum RepoLoaderError {
     #[error(transparent)]
-    TreeMerge(#[from] TreeMergeError),
+    Backend(#[from] BackendError),
     #[error(transparent)]
     OpHeadResolution(#[from] OpHeadResolutionError),
     #[error(transparent)]
@@ -670,13 +660,13 @@ impl RepoLoader {
             &self.op_store,
             |op_heads| self._resolve_op_heads(op_heads, user_settings),
         )?;
-        let view = View::new(op.view()?.take_store_view());
+        let view = op.view()?;
         Ok(self._finish_load(op, view))
     }
 
     #[instrument(skip(self))]
     pub fn load_at(&self, op: &Operation) -> Result<Arc<ReadonlyRepo>, RepoLoaderError> {
-        let view = View::new(op.view()?.take_store_view());
+        let view = op.view()?;
         Ok(self._finish_load(op.clone(), view))
     }
 
@@ -737,12 +727,27 @@ impl RepoLoader {
     }
 }
 
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+pub(crate) enum RewriteType {
+    Rewritten,
+    Divergent,
+    Abandoned,
+}
+
 pub struct MutableRepo {
     base_repo: Arc<ReadonlyRepo>,
     index: Box<dyn MutableIndex>,
     view: DirtyCell<View>,
-    rewritten_commits: HashMap<CommitId, HashSet<CommitId>>,
-    abandoned_commits: HashSet<CommitId>,
+    // TODO: make these fields private again
+    // The commit identified by the key has been replaced by all the ones in the value.
+    // * Branches pointing to the old commit should be updated to the new commit, resulting in a
+    //   conflict if there multiple new commits.
+    // * Children of the old commit should be rebased onto the new commits. However, if the type is
+    //   `Divergent`, they should be left in place.
+    // * Working copies pointing to the old commit should be updated to the first of the new
+    //   commits. However, if the type is `Abandoned`, a new working-copy commit should be created
+    //   on top of all of the new commits instead.
+    pub(crate) parent_mapping: HashMap<CommitId, (RewriteType, Vec<CommitId>)>,
 }
 
 impl MutableRepo {
@@ -757,8 +762,7 @@ impl MutableRepo {
             base_repo,
             index: mut_index,
             view: DirtyCell::with_clean(mut_view),
-            rewritten_commits: Default::default(),
-            abandoned_commits: Default::default(),
+            parent_mapping: Default::default(),
         }
     }
 
@@ -775,16 +779,15 @@ impl MutableRepo {
     }
 
     pub fn has_changes(&self) -> bool {
-        !(self.abandoned_commits.is_empty()
-            && self.rewritten_commits.is_empty()
-            && self.view() == &self.base_repo.view)
+        !(self.parent_mapping.is_empty() && self.view() == &self.base_repo.view)
     }
 
-    pub fn consume(self) -> (Box<dyn MutableIndex>, View) {
+    pub(crate) fn consume(self) -> (Box<dyn MutableIndex>, View) {
         self.view.ensure_clean(|v| self.enforce_view_invariants(v));
         (self.index, self.view.into_inner())
     }
 
+    /// Returns a [`CommitBuilder`] to write new commit to the repo.
     pub fn new_commit(
         &mut self,
         settings: &UserSettings,
@@ -794,6 +797,7 @@ impl MutableRepo {
         CommitBuilder::for_new_commit(self, settings, parents, tree_id)
     }
 
+    /// Returns a [`CommitBuilder`] to rewrite an existing commit in the repo.
     pub fn rewrite_commit(
         &mut self,
         settings: &UserSettings,
@@ -804,7 +808,8 @@ impl MutableRepo {
         // `self.rewritten_commits`
     }
 
-    pub fn write_commit(
+    /// Only called from [`CommitBuilder::write`]. Use that function instead.
+    pub(crate) fn write_commit(
         &mut self,
         commit: backend::Commit,
         sign_with: Option<&mut SigningFn>,
@@ -814,49 +819,35 @@ impl MutableRepo {
         Ok(commit)
     }
 
-    /// Record a commit as having been rewritten to one or more ids in this
+    /// Record a commit as having been rewritten to another commit in this
     /// transaction.
     ///
     /// This record is used by `rebase_descendants` to know which commits have
     /// children that need to be rebased, and where to rebase them to. See the
     /// docstring for `record_rewritten_commit` for details.
-    //  TODO(ilyagr): Make this API saner, e.g. make `self.rewritten_commits` public
-    // and make empty values correspond to abandoned commits.
-    pub fn set_rewritten_commit(
+    pub fn set_rewritten_commit(&mut self, old_id: CommitId, new_id: CommitId) {
+        assert_ne!(old_id, *self.store().root_commit_id());
+        self.parent_mapping
+            .insert(old_id, (RewriteType::Rewritten, vec![new_id]));
+    }
+
+    /// Record a commit as being rewritten into multiple other commits in this
+    /// transaction.
+    ///
+    /// A later call to `rebase_descendants()` will update branches pointing to
+    /// `old_id` be conflicted and pointing to all pf `new_ids`. Working copies
+    /// pointing to `old_id` will be updated to point to the first commit in
+    /// `new_ids``. Descendants of `old_id` will be left alone.
+    pub fn set_divergent_rewrite(
         &mut self,
         old_id: CommitId,
         new_ids: impl IntoIterator<Item = CommitId>,
     ) {
         assert_ne!(old_id, *self.store().root_commit_id());
-        self.rewritten_commits
-            .insert(old_id, new_ids.into_iter().collect());
-    }
-
-    /// Record a commit as having been rewritten in this transaction. If it was
-    /// already rewritten, mark it as divergent (unlike `set_rewritten_commit`)
-    ///
-    /// This record is used by `rebase_descendants` to know which commits have
-    /// children that need to be rebased, and where to rebase the children (as
-    /// well as branches) to.
-    ///
-    /// The `rebase_descendants` logic treats these records as follows:
-    ///
-    /// - If a commit is recorded as rewritten to a single commit, its
-    ///   descendants would be rebased to become descendants of `new_id`. Any
-    ///   branches at `old_id` are also moved to `new_id`.
-    /// - If a commit is recorded as rewritten to more than one commit, it is
-    ///   assumed to have become divergent. Its descendants are *not* rebased.
-    ///   However, the *branches* that were at `old_id` are moved to each of the
-    ///   new ids, and thus become conflicted.
-    ///
-    /// In neither case would `rebase_descendants` modify the `old_id` commit
-    /// itself.
-    pub fn record_rewritten_commit(&mut self, old_id: CommitId, new_id: CommitId) {
-        assert_ne!(old_id, *self.store().root_commit_id());
-        self.rewritten_commits
-            .entry(old_id)
-            .or_default()
-            .insert(new_id);
+        self.parent_mapping.insert(
+            old_id.clone(),
+            (RewriteType::Divergent, new_ids.into_iter().collect()),
+        );
     }
 
     /// Record a commit as having been abandoned in this transaction.
@@ -868,37 +859,107 @@ impl MutableRepo {
     /// The `rebase_descendants` logic will rebase the descendants of `old_id`
     /// to become the descendants of parent(s) of `old_id`. Any branches at
     /// `old_id` would be moved to the parent(s) of `old_id` as well.
+    // TODO: Propagate errors from commit lookup or take a Commit as argument.
     pub fn record_abandoned_commit(&mut self, old_id: CommitId) {
         assert_ne!(old_id, *self.store().root_commit_id());
-        self.abandoned_commits.insert(old_id);
+        // Descendants should be rebased onto the commit's parents
+        let old_commit = self.store().get_commit(&old_id).unwrap();
+        self.record_abandoned_commit_with_parents(old_id, old_commit.parent_ids().to_vec());
     }
 
-    fn clear_descendant_rebaser_plans(&mut self) {
-        self.rewritten_commits.clear();
-        self.abandoned_commits.clear();
+    /// Record a commit as having been abandoned in this transaction.
+    ///
+    /// A later `rebase_descendants()` will rebase children of `old_id` onto
+    /// `new_parent_ids`. A working copy pointing to `old_id` will point to a
+    /// new commit on top of `new_parent_ids`.
+    pub fn record_abandoned_commit_with_parents(
+        &mut self,
+        old_id: CommitId,
+        new_parent_ids: impl IntoIterator<Item = CommitId>,
+    ) {
+        assert_ne!(old_id, *self.store().root_commit_id());
+        self.parent_mapping.insert(
+            old_id,
+            (RewriteType::Abandoned, new_parent_ids.into_iter().collect()),
+        );
     }
 
     pub fn has_rewrites(&self) -> bool {
-        !(self.rewritten_commits.is_empty() && self.abandoned_commits.is_empty())
+        !self.parent_mapping.is_empty()
+    }
+
+    /// Calculates new parents for a commit that's currently based on the given
+    /// parents. It does that by considering how previous commits have been
+    /// rewritten and abandoned.
+    ///
+    /// Panics if `parent_mapping` contains cycles
+    pub fn new_parents(&self, old_ids: &[CommitId]) -> Vec<CommitId> {
+        fn single_substitution_round(
+            parent_mapping: &HashMap<CommitId, (RewriteType, Vec<CommitId>)>,
+            ids: Vec<CommitId>,
+        ) -> (Vec<CommitId>, bool) {
+            let mut made_replacements = false;
+            let mut new_ids = vec![];
+            // TODO(ilyagr): (Maybe?) optimize common case of replacements all
+            // being singletons. If CommitId-s were Copy. no allocations would be needed in
+            // that case, but it probably doesn't matter much while they are Vec<u8>-s.
+            for id in ids.into_iter() {
+                match parent_mapping.get(&id) {
+                    None | Some((RewriteType::Divergent, _)) => {
+                        new_ids.push(id);
+                    }
+                    Some((_, replacements)) => {
+                        assert!(
+                            // Each commit must have a parent, so a parent can
+                            // not just be mapped to nothing. This assertion
+                            // could be removed if this function is used for
+                            // mapping something other than a commit's parents.
+                            !replacements.is_empty(),
+                            "Found empty value for key {id:?} in the parent mapping",
+                        );
+                        made_replacements = true;
+                        new_ids.extend(replacements.iter().cloned())
+                    }
+                };
+            }
+            (new_ids, made_replacements)
+        }
+
+        let mut new_ids: Vec<CommitId> = old_ids.to_vec();
+        // The longest possible non-cycle substitution sequence goes through each key of
+        // parent_mapping once.
+        let mut allowed_iterations = 0..self.parent_mapping.len();
+        loop {
+            let made_replacements;
+            (new_ids, made_replacements) = single_substitution_round(&self.parent_mapping, new_ids);
+            if !made_replacements {
+                break;
+            }
+            allowed_iterations
+                .next()
+                .expect("cycle detected in the parent mapping");
+        }
+        match new_ids.as_slice() {
+            // The first two cases are an optimization for the common case of commits with <=2
+            // parents
+            [_singleton] => new_ids,
+            [a, b] if a != b => new_ids,
+            _ => new_ids.into_iter().unique().collect(),
+        }
     }
 
     /// After the rebaser returned by this function is dropped,
-    /// self.clear_descendant_rebaser_plans() needs to be called.
+    /// self.parent_mapping needs to be cleared.
     fn rebase_descendants_return_rebaser<'settings, 'repo>(
         &'repo mut self,
         settings: &'settings UserSettings,
         options: RebaseOptions,
-    ) -> Result<Option<DescendantRebaser<'settings, 'repo>>, TreeMergeError> {
+    ) -> BackendResult<Option<DescendantRebaser<'settings, 'repo>>> {
         if !self.has_rewrites() {
             // Optimization
             return Ok(None);
         }
-        let mut rebaser = DescendantRebaser::new(
-            settings,
-            self,
-            self.rewritten_commits.clone(),
-            self.abandoned_commits.clone(),
-        );
+        let mut rebaser = DescendantRebaser::new(settings, self);
         *rebaser.mut_options() = options;
         rebaser.rebase_all()?;
         Ok(Some(rebaser))
@@ -911,11 +972,11 @@ impl MutableRepo {
         &mut self,
         settings: &UserSettings,
         options: RebaseOptions,
-    ) -> Result<usize, TreeMergeError> {
+    ) -> BackendResult<usize> {
         let result = self
             .rebase_descendants_return_rebaser(settings, options)?
             .map_or(0, |rebaser| rebaser.into_map().len());
-        self.clear_descendant_rebaser_plans();
+        self.parent_mapping.clear();
         Ok(result)
     }
 
@@ -936,25 +997,25 @@ impl MutableRepo {
         &mut self,
         settings: &UserSettings,
         options: RebaseOptions,
-    ) -> Result<HashMap<CommitId, CommitId>, TreeMergeError> {
+    ) -> BackendResult<HashMap<CommitId, CommitId>> {
         let result = Ok(self
             // We do not set RebaseOptions here, since this function does not currently return
             // enough information to describe the results of a rebase if some commits got
             // abandoned
             .rebase_descendants_return_rebaser(settings, options)?
             .map_or(HashMap::new(), |rebaser| rebaser.into_map()));
-        self.clear_descendant_rebaser_plans();
+        self.parent_mapping.clear();
         result
     }
 
-    pub fn rebase_descendants(&mut self, settings: &UserSettings) -> Result<usize, TreeMergeError> {
+    pub fn rebase_descendants(&mut self, settings: &UserSettings) -> BackendResult<usize> {
         self.rebase_descendants_with_options(settings, Default::default())
     }
 
     pub fn rebase_descendants_return_map(
         &mut self,
         settings: &UserSettings,
-    ) -> Result<HashMap<CommitId, CommitId>, TreeMergeError> {
+    ) -> BackendResult<HashMap<CommitId, CommitId>> {
         self.rebase_descendants_with_options_return_map(settings, Default::default())
     }
 
@@ -1026,19 +1087,20 @@ impl MutableRepo {
 
     fn enforce_view_invariants(&self, view: &mut View) {
         let view = view.store_view_mut();
-        view.public_head_ids = self
-            .index()
-            .heads(&mut view.public_head_ids.iter())
-            .iter()
-            .cloned()
-            .collect();
-        view.head_ids.extend(view.public_head_ids.iter().cloned());
-        view.head_ids = self
-            .index()
-            .heads(&mut view.head_ids.iter())
-            .iter()
-            .cloned()
-            .collect();
+        let root_commit_id = self.store().root_commit_id();
+        if view.head_ids.is_empty() {
+            view.head_ids.insert(root_commit_id.clone());
+        } else if view.head_ids.len() > 1 {
+            // An empty head_ids set is padded with the root_commit_id, but the
+            // root id is unwanted during the heads resolution.
+            view.head_ids.remove(root_commit_id);
+            view.head_ids = self
+                .index()
+                .heads(&mut view.head_ids.iter())
+                .into_iter()
+                .collect();
+        }
+        assert!(!view.head_ids.is_empty());
     }
 
     /// Ensures that the given `head` and ancestor commits are reachable from
@@ -1107,16 +1169,6 @@ impl MutableRepo {
         self.view.mark_dirty();
     }
 
-    pub fn add_public_head(&mut self, head: &Commit) {
-        self.view_mut().add_public_head(head.id());
-        self.view.mark_dirty();
-    }
-
-    pub fn remove_public_head(&mut self, head: &CommitId) {
-        self.view_mut().remove_public_head(head);
-        self.view.mark_dirty();
-    }
-
     /// Returns true if any local or remote branch of the given `name` exists.
     #[must_use]
     pub fn has_branch(&self, name: &str) -> bool {
@@ -1284,13 +1336,6 @@ impl MutableRepo {
             }
         }
 
-        for removed_head in base.public_heads().difference(other.public_heads()) {
-            self.view_mut().remove_public_head(removed_head);
-        }
-        for added_head in other.public_heads().difference(base.public_heads()) {
-            self.view_mut().add_public_head(added_head);
-        }
-
         let base_heads = base.heads().iter().cloned().collect_vec();
         let own_heads = self.view().heads().iter().cloned().collect_vec();
         let other_heads = other.heads().iter().cloned().collect_vec();
@@ -1367,8 +1412,13 @@ impl MutableRepo {
             rewritten_changes.insert(change_id);
         }
         for (old_commit, new_commits) in rewritten_commits {
-            for new_commit in new_commits {
-                self.record_rewritten_commit(old_commit.clone(), new_commit);
+            if new_commits.len() == 1 {
+                self.set_rewritten_commit(
+                    old_commit.clone(),
+                    new_commits.into_iter().next().unwrap(),
+                );
+            } else {
+                self.set_divergent_rewrite(old_commit.clone(), new_commits);
             }
         }
 
@@ -1395,24 +1445,22 @@ impl Repo for MutableRepo {
         self.index.as_index()
     }
 
-    fn submodule_store(&self) -> &Arc<dyn SubmoduleStore> {
-        self.base_repo.submodule_store()
-    }
-
     fn view(&self) -> &View {
         self.view
             .get_or_ensure_clean(|v| self.enforce_view_invariants(v))
     }
 
+    fn submodule_store(&self) -> &Arc<dyn SubmoduleStore> {
+        self.base_repo.submodule_store()
+    }
+
     fn resolve_change_id_prefix(&self, prefix: &HexPrefix) -> PrefixResolution<Vec<CommitId>> {
-        let revset = RevsetExpression::all().evaluate_programmatic(self).unwrap();
-        let change_id_index = revset.change_id_index();
+        let change_id_index = self.index.change_id_index(&mut self.view().heads().iter());
         change_id_index.resolve_prefix(prefix)
     }
 
     fn shortest_unique_change_id_prefix_len(&self, target_id: &ChangeId) -> usize {
-        let revset = RevsetExpression::all().evaluate_programmatic(self).unwrap();
-        let change_id_index = revset.change_id_index();
+        let change_id_index = self.index.change_id_index(&mut self.view().heads().iter());
         change_id_index.shortest_unique_prefix_len(target_id)
     }
 }
@@ -1425,8 +1473,8 @@ pub struct RewriteRootCommit;
 /// Error from attempts to edit a commit
 #[derive(Debug, Error)]
 pub enum EditCommitError {
-    #[error("Current working-copy commit not found: {0}")]
-    WorkingCopyCommitNotFound(BackendError),
+    #[error("Current working-copy commit not found")]
+    WorkingCopyCommitNotFound(#[source] BackendError),
     #[error("Cannot rewrite the root commit")]
     RewriteRootCommit,
 }
@@ -1434,9 +1482,9 @@ pub enum EditCommitError {
 /// Error from attempts to check out a commit
 #[derive(Debug, Error)]
 pub enum CheckOutCommitError {
-    #[error("Failed to create new working-copy commit: {0}")]
+    #[error("Failed to create new working-copy commit")]
     CreateCommit(#[from] BackendError),
-    #[error("Failed to edit commit: {0}")]
+    #[error("Failed to edit commit")]
     EditCommit(#[from] EditCommitError),
 }
 
