@@ -29,7 +29,6 @@ use std::path::PathBuf;
 use std::sync::{Arc, RwLock};
 
 use blake2::{Blake2b512, Digest};
-use byteorder::{LittleEndian, ReadBytesExt, WriteBytesExt};
 use tempfile::NamedTempFile;
 use thiserror::Error;
 
@@ -74,7 +73,12 @@ impl ReadonlyTable {
         name: String,
         key_size: usize,
     ) -> TableStoreResult<Arc<ReadonlyTable>> {
-        let parent_filename_len = file.read_u32::<LittleEndian>()?;
+        let read_u32 = |file: &mut dyn Read| -> io::Result<u32> {
+            let mut buf = [0; 4];
+            file.read_exact(&mut buf)?;
+            Ok(u32::from_le_bytes(buf))
+        };
+        let parent_filename_len = read_u32(file)?;
         let maybe_parent_file = if parent_filename_len > 0 {
             let mut parent_filename_bytes = vec![0; parent_filename_len as usize];
             file.read_exact(&mut parent_filename_bytes)?;
@@ -84,7 +88,7 @@ impl ReadonlyTable {
         } else {
             None
         };
-        let num_local_entries = file.read_u32::<LittleEndian>()? as usize;
+        let num_local_entries = read_u32(file)? as usize;
         let index_size = num_local_entries * ReadonlyTableIndexEntry::size(key_size);
         let mut data = vec![];
         file.read_to_end(&mut data)?;
@@ -182,9 +186,7 @@ impl<'table> ReadonlyTableIndexEntry<'table> {
     }
 
     fn value_offset(&self) -> usize {
-        (&self.data[self.data.len() - 4..self.data.len()])
-            .read_u32::<LittleEndian>()
-            .unwrap() as usize
+        u32::from_le_bytes(self.data[self.data.len() - 4..].try_into().unwrap()) as usize
     }
 }
 
@@ -256,24 +258,22 @@ impl MutableTable {
         let mut buf = vec![];
 
         if let Some(parent_file) = &self.parent_file {
-            buf.write_u32::<LittleEndian>(parent_file.name.len() as u32)
-                .unwrap();
-            buf.write_all(parent_file.name.as_bytes()).unwrap();
+            buf.extend(u32::try_from(parent_file.name.len()).unwrap().to_le_bytes());
+            buf.extend_from_slice(parent_file.name.as_bytes());
         } else {
-            buf.write_u32::<LittleEndian>(0).unwrap();
+            buf.extend(0_u32.to_le_bytes());
         }
 
-        buf.write_u32::<LittleEndian>(self.entries.len() as u32)
-            .unwrap();
+        buf.extend(u32::try_from(self.entries.len()).unwrap().to_le_bytes());
 
-        let mut value_offset = 0;
+        let mut value_offset = 0_u32;
         for (key, value) in &self.entries {
-            buf.write_all(key).unwrap();
-            buf.write_u32::<LittleEndian>(value_offset).unwrap();
-            value_offset += value.len() as u32;
+            buf.extend_from_slice(key);
+            buf.extend(value_offset.to_le_bytes());
+            value_offset += u32::try_from(value.len()).unwrap();
         }
         for value in self.entries.values() {
-            buf.write_all(value).unwrap();
+            buf.extend_from_slice(value);
         }
         buf
     }
@@ -359,10 +359,7 @@ impl TableSegment for MutableTable {
 
 #[derive(Debug, Error)]
 #[error(transparent)]
-pub enum TableStoreError {
-    IoError(#[from] io::Error),
-    PersistError(#[from] tempfile::PersistError),
-}
+pub struct TableStoreError(#[from] pub io::Error);
 
 pub type TableStoreResult<T> = Result<T, TableStoreError>;
 
